#-*- coding: utf-8 -*-

from bgesdk.error import APIError

import pytest


class TestVariant:

    @pytest.mark.parametrize('rsids', ['rs333', 'rs1,rs2'])
    def test_valid_request(self, api, self_biosample_id, rsids):
        # 获取不归属于自己的样品的变异数据
        ret = api.get_variants(self_biosample_id, rsids)
        assert isinstance(ret, list), u'成功返回数据的类型必须为列表'

    @pytest.mark.parametrize('rsids', ['rs333', 'rs1,rs2'])
    def test_raise_errors(self, api, logger, other_biosample_id, rsids):
        """获取不归属于自己的样品的变异数据"""
        with pytest.raises(APIError) as e:
            api.get_variants(other_biosample_id, rsids)
        logger.debug(e.value)
        assert e.value.code == 41303
        assert e.value.msg == u'请提供本人的 biosample_id'

    @pytest.mark.parametrize(
        'rsids', ['rs1,rs2,rs3,rs4,rs5,rs6,rs7,rs8,rs9,rs10,rs11,rs12,rs13,'
                  'rs14,rs15,rs16,rs17,rs18,rs19,rs20,rs21,rs22,rs23,rs24,r'
                  's25,rs26,rs27,rs28,rs29,rs30,rs31,rs32,rs33,rs34,rs35,rs'
                  '36,rs37,rs38,rs39,rs40,rs41,rs42,rs43,rs44,rs45,rs46,rs4'
                  '7,rs48,rs49,rs50,rs51,rs52,rs53,rs54,rs55,rs56,rs57,rs58'
                  ',rs59,rs60,rs61,rs62,rs63,rs64,rs65,rs66,rs67,rs68,rs69,'
                  'rs70,rs71,rs72,rs73,rs74,rs75,rs76,rs77,rs78,rs79,rs80,r'
                  's81,rs82,rs83,rs84,rs85,rs86,rs87,rs88,rs89,rs90,rs91,rs'
                  '92,rs93,rs94,rs95,rs96,rs97,rs98,rs99,rs100,rs101'])
    def test_too_many_rsids(self, api, logger, self_biosample_id, rsids):
        """测试超出数量限制"""
        with pytest.raises(APIError) as e:
            api.get_variants(self_biosample_id, rsids)
        logger.debug(e.value)
        assert e.value.code == 41001
        assert e.value.msg == u'参数错误'

    @pytest.mark.parametrize(
        'rsids', ['333', '55333,rs3', 'srs3', 'rs3,,rs34', 'rs3,,', ''])
    def test_invalid_rsid(self, api, logger, self_biosample_id, rsids):
        """测试格式错误的变异位点"""
        with pytest.raises(APIError) as e:
            api.get_variants(self_biosample_id, rsids)
        logger.debug(e.value)
        assert e.value.code == 41001
        assert e.value.msg == u'参数错误'

    @pytest.mark.parametrize('biosample_id', ['demo', 'E-B112'])
    def test_invalid_biosample_id(self, api, logger, biosample_id):
        """测试不存在或者格式异常的的样品编号"""
        rsids = 'rs333'
        with pytest.raises(APIError) as e:
            api.get_variants(biosample_id, rsids)
        logger.debug(e.value)
        assert e.value.code == 41302
        assert e.value.msg == u'基因组报告不存在或未下机'



import os

from posixpath import splitext

from bgesdk.management.command import BaseCommand
from bgesdk.management.utils import (
    get_config_dir, get_config_path, get_active_path, get_active_project,
    output
)


class Command(BaseCommand):

    order = 2
    help='激活某项目配置，其他命令均使用已激活的项目配置作为全局配置。'

    def add_arguments(self, parser):
        parser.add_argument(
            'project',
            nargs='?',
            help='自定义项目名称保存全局配置，其他命令默认读取正在生效的项目配置。'
        )

    def handler(self, args):
        """将配置写入项目文件"""
        project = args.project
        if not project:
            return self.list_projects()
        else:
            project = project.lower()
            return self.workon_project(project)

    def workon_project(self, project):
        config_path = get_config_path(project)
        output('配置文件位置：{}'.format(config_path))
        active_path = get_active_path()
        with open(active_path, 'w') as fp:
            fp.write(project)
        output('已激活 {} 的项目配置'.format(project))

    def list_projects(self):
        active_project = get_active_project()
        config_dir = get_config_dir()
        projects = []
        for filename in os.listdir(config_dir):
            name, ext = splitext(filename)
            if ext != '.ini':
                continue
            projects.append(name)
        output('通过 bge workon <NAME> 切换生效配置：\n')
        projects.sort()
        if active_project in projects:
            output(active_project, '- 已生效')
        for project in projects:
            if project == active_project:
                continue
            output(project)



#-*- coding: utf-8 -*-

from bgesdk.client import OAuth2

import logging
import pytest
import os


ENDPOINT = os.environ.get('ENDPOINT')
CLIENT_ID = os.environ.get('CLIENT_ID')
CLIENT_SECRET = os.environ.get('CLIENT_SECRET')
REDIRECT_URL = os.environ.get('REDIRECT_URL')
ACCESS_TOKEN = os.environ.get('ACCESS_TOKEN')
SELF_BIOSAMPLE_ID = os.environ.get('SELF_BIOSAMPLE_ID')
SELF_META_BIOSAMPLE_ID = os.environ.get('SELF_META_BIOSAMPLE_ID')
OTHER_BIOSAMPLE_ID = os.environ.get('OTHER_BIOSAMPLE_ID')


@pytest.fixture(scope='session')
def redirect_url():
    return REDIRECT_URL


@pytest.fixture(scope='session')
def logger():
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)
    return logger


@pytest.fixture(scope='session')
def oauth2():
    return OAuth2(CLIENT_ID, CLIENT_SECRET, endpoint=ENDPOINT)


@pytest.fixture(scope='session')
def access_token():
    return ACCESS_TOKEN


@pytest.fixture(scope='session')
def api(oauth2, access_token):
    return oauth2.get_api(access_token)


@pytest.fixture(scope='session')
def self_biosample_id():
    """本人的样品编号"""
    return SELF_BIOSAMPLE_ID


@pytest.fixture(scope='session')
def self_meta_biosample_id():
    """本人的 Meta 样品编号"""
    return SELF_META_BIOSAMPLE_ID


@pytest.fixture(scope='session')
def other_biosample_id():
    """其他人的样品编号"""
    return OTHER_BIOSAMPLE_ID




# For terminal commands
DEFAULT_MODEL_SECTION = 'Model'
DEFAULT_OAUTH2_SECTION = 'OAuth2'
DEFAULT_TOKEN_SECTION = 'Token'
DEFAULT_PROJECT = 'default'
DEFAULT_MODEL_TIMEOUT = None

MODEL_CONTAINER_PREFIX = 'bge-model-'
TEST_SERVER_PORT = 9999
TEST_SERVER_ENDPOINT = 'http://0.0.0.0:{}'.format(TEST_SERVER_PORT)


CLIENT_CREDENTIALS_CONFIGS = [
    ('client_id', {
        'type': 'str',
        'description': '客户端编号',
        'secure': True
    }),
    ('client_secret', {
        'type': 'str',
        'description': '客户端密钥',
        'secure': True
    }),
    ('endpoint', {
        'default': 'https://api.bge.genomics.cn',
        'type': 'str',
        'description': '访问域名'
    })
]

TAB_CHOICES = [
        ('genetic_diseases', '遗传疾病'),
        ('health_risks', '健康风险'),
        ('physiological_index', '生理指标'),
        ('comprehensive', '综合'),
        ('tools', '工具')
    ]
LANGUAGE_CHOICES = [('zh', '中文版'), ('en', 'English')]

TITLE_NAME = {
    'brief_intro': '简介',
    'method': '方法',
    'model_evaluation': '模型评价',
    'data_set_size': '数据规模',
    'ethnicity': '研究族群',
    'limitation': '局限性说明',
}


API_TABLE = {
    'QueryParams': '请求参数',
    'Success': '返回参数',
    'State': '参数'
}



import sys

from bgesdk.client import OAuth2
from bgesdk.error import APIError
from bgesdk.management import constants
from bgesdk.management.command import BaseCommand
from bgesdk.management.utils import (
    get_active_project, config_get, get_config_path, get_config_parser,
    output
)


class Command(BaseCommand):

    order = 4
    help='获取、保存访问令牌。'

    def handler(self, args):
        project = get_active_project()
        config_path = get_config_path(project)
        config_parser = get_config_parser(config_path)
        oauth2_section = constants.DEFAULT_OAUTH2_SECTION
        client_id = config_get(config_parser.get, oauth2_section, 'client_id')
        client_secret = config_get(
            config_parser.get, oauth2_section, 'client_secret')
        endpoint = config_get(config_parser.get, oauth2_section, 'endpoint')
        oauth2 = OAuth2(
            client_id, client_secret, endpoint=endpoint, timeout=60.)
        try:
            token_result = oauth2.get_credentials_token()
        except APIError as e:
            output('令牌获取出错: {}'.format(e))
            sys.exit(1)
        self._write_token_config(project, token_result)
        output('令牌内容如下：')
        output('')
        for key in ['access_token', 'token_type', 'expires_in', 'scope']:
            output('{} = {}'.format(key, token_result[key]))

    def _write_token_config(self, project, token_result):
        access_token = token_result['access_token']
        token_type = token_result['token_type']
        expires_in = str(token_result['expires_in'])
        scope = token_result['scope']
        config_path = get_config_path(project)
        config_parser = get_config_parser(config_path)
        section_name = constants.DEFAULT_TOKEN_SECTION
        if section_name not in config_parser.sections():
            config_parser.add_section(section_name)
        config_parser.set(section_name, 'access_token', access_token)
        config_parser.set(section_name, 'token_type', token_type)
        config_parser.set(section_name, 'expires_in', expires_in)
        config_parser.set(section_name, 'scope', scope)
        with open(config_path, 'w') as config_file:
            config_parser.write(config_file)
        output('令牌已保存至：{}'.format(config_path))



import argparse

from abc import ABCMeta, abstractmethod


class BaseCommand(metaclass=ABCMeta):

    help = ''

    def __init__(self, parser, command_name=None, **kwargs):
        if command_name is not None:
            parser = parser.add_parser(
                command_name,
                formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                help=self.help or '',
                **kwargs
            )
        parser.set_defaults(method=self.handler, parser=parser)
        self.parser = parser
        self.add_arguments(parser)

    def add_arguments(self, parser):
        pass

    @abstractmethod
    def handler(self, args):
        pass



#-*- coding: utf-8 -*-

import requests

from requests.adapters import HTTPAdapter
from six.moves.urllib.parse import urljoin

from .error import APIError, BGEError
from .utils import new_logger


class HTTPRequest(object):
    """HTTP 请求类

    处理 SDK 接口的 HTTP 调用；

    Args:
        endpoint (str): 平台基本地址，如 https://api.bge.genomics.cn；
        max_retries (数字, 非必填): 接口请求重试次数，默认值为 None；
        verbose (布尔, 非必填)：输出测试日志，默认值为 False；
    """

    def __init__(self, endpoint, max_retries=None, verbose=False):
        self.logger = new_logger(self.__class__.__name__, verbose=verbose)
        self.endpoint = endpoint
        self.headers = {}
        self.session = requests.Session()
        if max_retries is not None:
            self.session.mount(
                'http://', HTTPAdapter(max_retries=max_retries))
            self.session.mount(
                'https://', HTTPAdapter(max_retries=max_retries))

    def set_authorization(self, token_type, access_token):
        """设置 Authorization 头部

        Args:
            access_token (str): 访问令牌
        """
        self.headers['Authorization'] = '{} {}'.format(
            token_type, access_token)

    def get(self, path, params=None, headers=None, timeout=None):
        """GET 接口请求

        Args:
            path (str): 请求路径
            params (dict, 非必填): GET 参数；
            headers (dict, 非必填): HTTP 头部；
            timeout (int, 非必填): 请求超时时间；

        Returns:
            object: 请求返回值
        """
        return self._request(
            'GET', path, params=params, headers=headers, timeout=timeout)

    def post(self, path, params=None, data=None, files=None, headers=None,
             timeout=None):
        """POST

        Args:
            path (str): 请求路径
            params (dict, 非必填): GET 参数；
            data (dict, 非必填): POST 参数；
            files (dict, 非必填): 文件参数；
            headers (dict, 非必填): HTTP 头部；
            timeout (int, 非必填): 请求超时时间；

        Returns:
            object: 请求返回值
        """
        return self._request(
            'POST', path, params=params, data=data, files=files,
            headers=headers, timeout=timeout)

    def _request(self, method, path, timeout=None, headers=None, **kwargs):
        """发送 HTTP 请求

        Args:
            method (str): 请求方法
            path (str): 请求路径

        Raises:
            BGEError: SDK 错误
            APIError: API 错误

        Returns:
            object: 返回值
        """
        if headers is None:
            headers = self.headers
        else:
            headers.update(self.headers)
        url = urljoin(self.endpoint, path)
        self.logger.debug(
            ('Request: \n\tmethod: %s\n\turl: %s\n\theaders: %s\n\ttimeout'
             ': %s\n\t**kwargs=%s'), method, url, headers, timeout, kwargs)
        try:
            resp = self.session.request(
                method=method, url=url, headers=headers, timeout=timeout,
                **kwargs)
            resp.raise_for_status()
        except requests.exceptions.HTTPError as e:
            raise BGEError('API request error: %s' % e)
        except requests.exceptions.ConnectionError as e:
            raise BGEError('Fail to connect: %s' % e)
        content_type_header = resp.headers.get('Content-Type', '')
        if 'application/json' not in content_type_header:
            result = resp.text
            self.logger.debug('Response: \n\t%s', result)
            return result
        result = resp.json()
        self.logger.debug('Response: \n\t%s', result)
        data = result.get('data')
        code = result['code']
        msg = result['msg']
        if code != 0:
            raise APIError(code, msg, data)
        pagination = result.get('pagination')
        if pagination:
            # 相关接口分页返回方式升级后，可去除此处
            # TODO: upgrade in the future
            return data, pagination
        return data



#-*- coding: utf-8 -*-

from __future__ import unicode_literals

import json


class BGEError(Exception):
    """SDK 错误"""


class APIError(BGEError):
    """接口错误"""

    def __init__(self, code, msg, data=None):
        self.code = int(code)
        self.msg = msg
        self.data = data
        super(APIError, self).__init__({
            'code': self.code,
            'msg': self.msg,
            'data': self.data
        })



import argparse
import os
import sys

from posixpath import exists
from six.moves import input

from bgesdk.management import constants
from bgesdk.management.command import BaseCommand
from bgesdk.management.utils import (
    get_config_parser, config_get, secure_str, get_config_path, confirm,
    get_active_project, output
)


class Command(BaseCommand):

    order = 1
    help = '配置、新增、显示、删除 BGE 开放平台的 OAuth2 配置，仅支持客户端模式。'

    def add_arguments(self, parser):
        sub_ps = parser.add_subparsers(
            help='新增、删除、显示 OAuth2 配置',
            required=False
        )
        add_p = sub_ps.add_parser(
            'add',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            help='新增 BGE 开放平台配置。'
        )
        add_p.add_argument(
            'project',
            help='全局配置项目名。'
        )
        add_p.set_defaults(method=self.add_project, parser=add_p)
        remove_p = sub_ps.add_parser(
            'remove',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            help='配置、显示、删除 BGE 开放平台的 OAuth2 全局配置，仅支持客户端模式。'
        )
        remove_p.add_argument(
            'project',
            help='全局配置项目名。'
        )
        remove_p.set_defaults(method=self.remove_project, parser=remove_p)
        show_p = sub_ps.add_parser(
            'show',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            help='打印显示全局配置信息。'
        )
        show_p.add_argument(
            'project',
            nargs='?',
            help='全局配置项目名。'
        )
        show_p.set_defaults(method=self.show_project, parser=show_p)

    def handler(self, args):
        project = get_active_project()
        output('当前正在配置 {}：\n'.format(project))
        config_path = get_config_path(project, check_exists=False)
        config_parser = get_config_parser(config_path)
        oauth2_section = constants.DEFAULT_OAUTH2_SECTION
        if oauth2_section not in config_parser.sections():
            config_parser.add_section(oauth2_section)
        for key, conf in constants.CLIENT_CREDENTIALS_CONFIGS:
            type_ = conf['type']
            if 'str' == type_:
                saved_value = config_get(
                    config_parser.get, oauth2_section, key)
            elif 'int' == type_:
                saved_value = config_get(
                    config_parser.getint, oauth2_section, key)
            elif 'bool' == type_:
                saved_value = config_get(
                    config_parser.getboolean, oauth2_section, key)
            else:
                raise ValueError('invalid type: {}'.format(type_))
            secure = conf.get('secure')
            description = conf.get('description', '')
            value = saved_value or conf.get('default', '')
            if secure and value:
                value = secure_str(value)
            input_value = input(
                '？请输入{} {} [{}]：'.format(description, key, value))
            if input_value:
                config_parser.set(oauth2_section, key, input_value)
            elif saved_value is None:
                if conf.get('default') is not None:
                    config_parser.set(oauth2_section, key, conf.get('default'))
                else:
                    config_parser.set(oauth2_section, key, '')
        with open(config_path, 'w') as config_file:
            config_parser.write(config_file)
        output('')
        output('配置已保存至：{}'.format(config_path))

    def add_project(self, args):
        project = args.project.lower()
        output('正在添加配置 {}：\n'.format(project))
        config_path = get_config_path(project, check_exists=False)
        if exists(config_path):
            output('配置 {} 已存在'.format(project))
            sys.exit(1)
        config_parser = get_config_parser(config_path)
        oauth2_section = constants.DEFAULT_OAUTH2_SECTION
        if oauth2_section not in config_parser.sections():
            config_parser.add_section(oauth2_section)
        for key, conf in constants.CLIENT_CREDENTIALS_CONFIGS:
            type_ = conf['type']
            if 'str' == type_:
                saved_value = config_get(
                    config_parser.get, oauth2_section, key)
            elif 'int' == type_:
                saved_value = config_get(
                    config_parser.getint, oauth2_section, key)
            elif 'bool' == type_:
                saved_value = config_get(
                    config_parser.getboolean, oauth2_section, key)
            else:
                raise ValueError('invalid type: {}'.format(type_))
            secure = conf.get('secure')
            description = conf.get('description', '')
            value = saved_value or conf.get('default', '')
            if secure and value:
                value = secure_str(value)
            input_value = input(
                '？请输入{} {} [{}]：'.format(description, key, value))
            if input_value:
                config_parser.set(oauth2_section, key, input_value)
            elif saved_value is None:
                if conf.get('default') is not None:
                    config_parser.set(oauth2_section, key, conf.get('default'))
                else:
                    config_parser.set(oauth2_section, key, '')
        with open(config_path, 'w') as config_file:
            config_parser.write(config_file)
        output('')
        output('配置已保存至：{}'.format(config_path))

    def remove_project(self, args):
        project = args.project.lower()
        activate_project = get_active_project()
        if activate_project == project:
            output('无法删除正在使用的配置，删除前请先使用 bge workon project 切换')
            sys.exit(1)
        config_path = get_config_path(project)
        if not confirm(prompt='确认删除配置项目 {}？'.format(project)):
            output('已取消删除')
            sys.exit(0)
        try:
            os.unlink(config_path)
        except (IOError, OSError):
            pass
        output('成功删除配置项目 {}'.format(project))

    def show_project(self, args):
        project = args.project
        if not project:
            project = get_active_project()
        project = project.lower()
        config_path = get_config_path(project)
        output('配置文件：{}'.format(config_path))
        output('配置详情：')
        output('')
        with open(config_path, 'r') as config_file:
            output(config_file.read())



import posixpath
import sys

from importlib import import_module

from bgesdk.management import constants
from bgesdk.management.command import BaseCommand
from bgesdk.management.utils import find_commands
from bgesdk.version import __version__


DEFAULT_OAUTH2_SECTION = constants.DEFAULT_OAUTH2_SECTION
DEFAULT_TOKEN_SECTION = constants.DEFAULT_TOKEN_SECTION

COMMAND_NAME = 'api'


class Command(BaseCommand):

    order = 5
    help='BGE 开放平台接口测试工具，可调用部分平台接口。'

    def add_arguments(self, parser):
        api_subparsers = parser.add_subparsers(
            dest='subcommand',
            help='可选子命令。'
        )
        commands_dir = posixpath.join(__path__[0], 'commands')
        command_names = find_commands(commands_dir)
        sort_items = []
        for sub_command_name in command_names:
            module = import_module(
                'bgesdk.management.commands.{}.commands.{}'.format(
                    COMMAND_NAME, sub_command_name
                )
            )
            command_klass = getattr(module, 'Command', None)
            if command_klass and not issubclass(command_klass, BaseCommand):
                continue
            order = int(getattr(command_klass, 'order', 1))
            sort_items.append((order, sub_command_name, command_klass))
        sort_items.sort()
        for _, sub_command_name, command_klass in sort_items:
            command_klass(api_subparsers, sub_command_name)

    def handler(self, args):
        """打印 subparser 帮助信息"""
        parser = args.parser
        parser.print_help(sys.stderr)


import json
import sys

from bgesdk.client import API
from bgesdk.error import APIError
from bgesdk.management import constants
from bgesdk.management.command import BaseCommand
from bgesdk.management.utils import (
    get_active_project, config_get, read_config, output
)
from bgesdk.models import ModelEncoder
from bgesdk.version import __version__


DEFAULT_OAUTH2_SECTION = constants.DEFAULT_OAUTH2_SECTION
DEFAULT_TOKEN_SECTION = constants.DEFAULT_TOKEN_SECTION


class Command(BaseCommand):

    order = 3
    help='获取变异位点数据。'

    def add_arguments(self, parser):
        parser.add_argument(
            'biosample_id',
            type=str,
            help='生物样品编号。'
        )
        parser.add_argument(
            'rsids',
            type=str,
            help='变异位点，多个使用逗号分割。'
        )
        parser.add_argument(
            '--pretty',
            default=False,
            action='store_true',
            help='打印可读性高的 JSON 字符串。'
        )
        parser.add_argument(
            '-t',
            '--access_token',
            type=str,
            help='访问令牌，为空时默认使用 bge token 命令获取到的 access_token。'
        )

    def handler(self, args):
        pretty = args.pretty
        access_token = args.access_token
        biosample_id = args.biosample_id.upper()
        rsids = args.rsids
        project = get_active_project()
        oauth2_section = DEFAULT_OAUTH2_SECTION
        token_section = DEFAULT_TOKEN_SECTION
        config = read_config(project)
        if not access_token:
            access_token = config_get(config.get, token_section, 'access_token')
        endpoint = config_get(config.get, oauth2_section, 'endpoint')
        api = API(access_token, endpoint=endpoint, timeout=18.)
        try:
            result = api.get_variants(biosample_id, rsids)
        except APIError as e:
            output('请求失败：{}'.format(e))
            sys.exit(1)
        if not result:
            output('请求成功，返回值：')
            output(result)
            sys.exit(1)
        if pretty:
            result = json.dumps(
                result, ensure_ascii=False, indent=4, cls=ModelEncoder
            )
        else:
            result = json.dumps(result, ensure_ascii=False, cls=ModelEncoder)
        output('请求成功，返回值：')
        output(result)



from validator import Required, In, validate, Length

from bgesdk.management.constants import TAB_CHOICES, LANGUAGE_CHOICES

language = [x[0] for x in LANGUAGE_CHOICES]
tab = [x[0] for x in TAB_CHOICES]


def validator_doc(doc_data):
    results = {}
    result_errors = {}
    validation = {
        'doc_tab': [Required, In(tab)],
        'model_id': [
            Required, Length(0, maximum=50), lambda x: isinstance(x, str)],
        'doc_content': [Required, lambda x: isinstance(x, list)]
    }
    res = validate(validation, doc_data)
    result_errors.update(res.errors)
    if isinstance(doc_data['doc_content'], list):
        for content in doc_data['doc_content']:
            validation = {
                'language': [Required, In(language)],
                'doc_name': [
                    Required, Length(0, maximum=50),
                    lambda x: isinstance(x, str)],
                'content_title': [
                    Required, Length(0, maximum=200),
                    lambda x: isinstance(x, str)],
                'developer': [
                    Length(0, maximum=50), lambda x: isinstance(x, str)],
                'brief_intro': [lambda x: isinstance(x, dict)],
                'method': [lambda x: isinstance(x, dict)],
                'model_evaluation': [lambda x: isinstance(x, dict)],
                'data_set_size': [lambda x: isinstance(x, dict)],
                'ethnicity': [lambda x: isinstance(x, dict)],
                'limitation': [lambda x: isinstance(x, dict)],
                'return_params': [lambda x: isinstance(x, dict)],
                'state_explain': [lambda x: isinstance(x, dict)],
                'example_result': [lambda x: isinstance(x, dict)],
                'ref': [lambda x: isinstance(x, list)]
            }
            content_res = validate(validation, content)
            result_errors.update(content_res.errors)
    valid = True
    if result_errors:
        valid = False
    results['errors'] = result_errors
    results['valid'] = valid
    return results


#-*- coding: utf-8 -*-

# 支持的 OAuth2 授权方式
GRANT_AUTHORIZATION_CODE = 'authorization_code'

GRANT_TYPE_CREDENTIALS = 'client_credentials'

DEFAULT_ENDPOINT = 'https://api.bge.genomics.cn'

ENDPOINTS = [
    {
        'name': '线上环境主服务',
        'endpoint': DEFAULT_ENDPOINT
    },
    {
        'name': '线上环境备份域名',
        'endpoint': 'https://api.bge.omgut.com'
    }
]



import posixpath
import sys

from bgesdk.client import API
from bgesdk.error import APIError
from bgesdk.management import constants
from bgesdk.management.command import BaseCommand
from bgesdk.management.utils import (
    get_active_project, config_get, read_config, output
)
from bgesdk.version import __version__


DEFAULT_OAUTH2_SECTION = constants.DEFAULT_OAUTH2_SECTION
DEFAULT_TOKEN_SECTION = constants.DEFAULT_TOKEN_SECTION


class Command(BaseCommand):

    order = 10
    help='上传文件。'

    def add_arguments(self, parser):
        parser.add_argument(
            'filepath',
            type=str,
            help='要上传的文件路径。'
        )
        parser.add_argument(
            '-f',
            '--filename',
            help='服务器存储使用的文件名，不提供时默认使用上传文件的名字。'
        )
        parser.add_argument(
            '-i',
            '--cmk_id',
            help='阿里云 KMS 服务密钥 ID，提供时代表使用加密方式上传文件。'
        )
        parser.add_argument(
            '-t',
            '--access_token',
            type=str,
            help='访问令牌，为空时默认使用 bge token 命令获取到的 access_token。'
        )

    def handler(self, args):
        access_token = args.access_token
        filepath = args.filepath
        project = get_active_project()
        oauth2_section = DEFAULT_OAUTH2_SECTION
        token_section = DEFAULT_TOKEN_SECTION
        config = read_config(project)
        if not access_token:
            access_token = config_get(config.get, token_section, 'access_token')
        endpoint = config_get(config.get, oauth2_section, 'endpoint')
        api = API(access_token, endpoint=endpoint, timeout=18.)
        if not posixpath.isfile(filepath):
            output('只能上传文件')
            sys.exit(1)
        filename = args.filename
        cmk_id = args.cmk_id
        if not filename:
            filename = posixpath.split(filepath)[1]
        try:
            with open(filepath, 'rb') as fp:
                object_name = api.upload(filename, fp, cmk_id=cmk_id)
        except APIError as e:
            output('\n\n请求失败：{}'.format(e))
            sys.exit(1)
        output('\n\n文件上传成功：{}'.format(object_name))



import json
import sys

from bgesdk.client import API
from bgesdk.error import APIError
from bgesdk.management import constants
from bgesdk.management.command import BaseCommand
from bgesdk.management.utils import (
    get_active_project, config_get, read_config, output
)
from bgesdk.models import ModelEncoder
from bgesdk.version import __version__


DEFAULT_OAUTH2_SECTION = constants.DEFAULT_OAUTH2_SECTION
DEFAULT_TOKEN_SECTION = constants.DEFAULT_TOKEN_SECTION

NOT_PARAM_FIELDS = (
    'command', 'parser', 'method', 'pretty', 'access_token', 'subcommand'
)


class Command(BaseCommand):

    order = 2
    help='获取样品信息。'

    def add_arguments(self, parser):
        parser.add_argument(
            '-b',
            '--biosample_ids',
            type=str,
            help='生物样品编号，逗号分割多个。'
        )
        parser.add_argument(
            '--omics',
            type=str,
            help='所属组学，取值范围：1-2。'
        )
        parser.add_argument(
            '--project_ids',
            type=str,
            help='所属项目，逗号分割多个。'
        )
        parser.add_argument(
            '--organisms',
            type=str,
            help='样品生物体，取值范围：1-3。'
        )
        parser.add_argument(
            '--data_availability',
            default=None,
            action='store_true',
            help='数据可用。'
        )
        parser.add_argument(
            '--no-data_availability',
            dest='data_availability', 
            action='store_false',
            help='数据不可用。'
        )
        parser.add_argument(
            '--statuses',
            type=str,
            help='数据状态，详情见 BGE 开放平台文档 https://api.bge.genomics.cn/doc/#/ot'
                'hers/appendix?id=样品状态编码表'
        )
        parser.add_argument(
            '-p',
            '--next_page',
            type=int,
            help='要获取的页码。'
        )
        parser.add_argument(
            '-n',
            '--limit',
            type=int,
            help='每页返回数量，默认值为 50。'
        )
        parser.add_argument(
            '--pretty',
            default=False,
            action='store_true',
            help='打印可读性高的 JSON 字符串。'
        )
        parser.add_argument(
            '-t',
            '--access_token',
            type=str,
            help='访问令牌，为空时默认使用 bge token 命令获取到的 access_token。'
        )

    def handler(self, args):
        pretty = args.pretty
        access_token = args.access_token
        params = vars(args)
        for field in NOT_PARAM_FIELDS:
            params.pop(field, None)
        project = get_active_project()
        oauth2_section = DEFAULT_OAUTH2_SECTION
        token_section = DEFAULT_TOKEN_SECTION
        config = read_config(project)
        if not access_token:
            access_token = config_get(config.get, token_section, 'access_token')
        endpoint = config_get(config.get, oauth2_section, 'endpoint')
        api = API(access_token, endpoint=endpoint, timeout=18.)
        try:
            result = api.get_samples(**params)
        except APIError as e:
            output('请求失败：{}'.format(e))
            sys.exit(1)
        if pretty:
            result = json.dumps(
                result, ensure_ascii=False, indent=4, cls=ModelEncoder
            )
        else:
            result = json.dumps(result, ensure_ascii=False, cls=ModelEncoder)
        output('请求成功，返回值：')
        output(result)



#-*- coding: utf-8 -*-

"""
BGE 开放平台 SDK 客户端模块。

当前模块提供了两个主要的类 OAuth2 和 API，他们分别提供了对 BGE 开放平台的 OAuth2 相关
接口和其他接口的调用方法。

使用示例：

    >>> oauth2 = OAuth2(
            'demo', 'demo', endpoint='https://api.bge.genomics.cn')
    >>> token = oauth2.get_credentials_token()
    >>> api = oauth2.get_api(token.access_token)
    >>> api.invoke_model('demo_model_id')
    Model({...})
"""

from . import constants
from . import models
from .error import BGEError
from .fs import FileItem
from .http import HTTPRequest
from .utils import new_logger, human_byte

from aliyunsdkcore.auth.credentials import StsTokenCredential
from aliyunsdkcore.client import AcsClient
from posixpath import split, join, isdir
from requests_toolbelt.multipart import encoder
from six import text_type
from six.moves.urllib.parse import urljoin, urlencode

import json
import os
import oss2
import requests
import sys


__all__ = ['OAuth2', 'API', 'endpoints']

endpoints = [v['endpoint'] for v in constants.ENDPOINTS]


def progress_callback(bytes_consumed, total_bytes):
    sys.stdout.write(
        '\r文件大小：{}, 上传进度: {}%，已上传 {}'.format(
            human_byte(total_bytes, 2),
            '%.2f' % ((bytes_consumed / total_bytes) * 100),
            human_byte(bytes_consumed, 2)
    ))
    sys.stdout.flush()


def alive(self):
    """检查服务可用性

    Returns:
        布尔型: True 代表可用，False 代表不可用
    """
    timeout = self.timeout
    verbose = self.verbose
    max_retries = self.max_retries
    request = HTTPRequest(
        self.endpoint, max_retries=max_retries, verbose=verbose)
    try:
        request.get('/ping', timeout=timeout)
    except Exception as e:
        return False
    return True


class OAuth2(object):
    """OAuth2 授权客户端类。
    
    管理关于 OAuth2 相关接口的调用，包括获取授权页面地址、授权码交换访问令牌等；
    
    Args:
        client_id (字符串): 第三方客户端 client_id；
        client_secret (字符串): 第三方客户端 client_secret；
        endpoint (字符串, 非必填): 平台对外服务的访问域名，
                                 默认值为 https://api.bge.genomics.cn；
        max_retries (数字, 非必填): 接口请求重试次数，默认值为 3；
        timeout (数字, 非必填): 接口请求默认超时间，默认值为 None；
        verbose (布尔, 非必填)：输出测试日志，默认值为 False；
    """

    alive = alive

    def __init__(self, client_id, client_secret, endpoint=None,
                 max_retries=None, timeout=None, verbose=False):
        self.client_id = client_id
        self.client_secret = client_secret
        if endpoint is None:
            endpoint = constants.DEFAULT_ENDPOINT
        self.endpoint = endpoint
        if max_retries is not None:
            max_retries = int(max_retries)
        self.max_retries = max_retries
        if timeout is not None:
            timeout = int(timeout)
        self.timeout = timeout
        self.verbose = verbose
        self.logger = new_logger(self.__class__.__name__, verbose=verbose)

    def get_authorization_url(self, redirect_uri, state=None, scopes=None):
        """获取用户授权页链接地址。

        Args:
            redirect_uri (str): 回调地址；
            state (str, 非必填): 第三方自定义信息，返回授权码时原样返回，
                                默认值为 None；
            scopes (list, 非必填): 权限范围，支持多个。默认值为 None；

        Returns:
            str: 用户授权页面地址；
        """
        params = {
            "client_id": self.client_id,
            "response_type": "code",
            "redirect_uri": redirect_uri
        }
        if scopes is not None:
            params['scope'] = " ".join(scopes)
        if state is not None:
            params["state"] = state
        qs = urlencode(params)
        return '?'.join((urljoin(self.endpoint, '/oauth2/authorize'), qs))

    def exchange_authorization_code(self, code, redirect_uri):
        """用户授权码交换访问令牌

        Args:
            code (str): 用户授权后平台返回的授权码；
            redirect_uri (str): 回调地址；

        Returns:
            AuthorizationCodeToken: 与授权用户关联的访问令牌，同时包含有刷新令牌、
                                    过期时间等信息；
        """
        data = {
            'client_id': self.client_id,
            'client_secret': self.client_secret,
            'grant_type': constants.GRANT_AUTHORIZATION_CODE,
            'redirect_uri': redirect_uri,
            'code': code
        }
        timeout = self.timeout
        verbose = self.verbose
        max_retries = self.max_retries
        request = HTTPRequest(
            self.endpoint, max_retries=max_retries, verbose=verbose)
        result = request.post(
            '/oauth2/access_token', data=data, timeout=timeout)
        return models.AuthorizationCodeToken(self, result)

    def exchange_refresh_token(self, refresh_token):
        """刷新令牌 access_token

        Args:
            refresh_token (str): 授权码模式所获得的 refresh_token；

        Returns:
            Model: 新的令牌数据；
        """
        data = {
            'client_id': self.client_id,
            'client_secret': self.client_secret,
            'grant_type': 'refresh_token',
            'refresh_token': refresh_token
        }
        timeout = self.timeout
        verbose = self.verbose
        max_retries = self.max_retries
        request = HTTPRequest(
            self.endpoint, max_retries=max_retries, verbose=verbose)
        result = request.post(
            '/oauth2/access_token', data=data, timeout=timeout)
        return models.AuthorizationCodeToken(self, result)

    def get_credentials_token(self):
        """客户端授权模式下获取访问令牌

        Returns:
            Model: 访问令牌，包含 access_token、过期时间等；
        """
        data = {
            'client_id': self.client_id,
            'client_secret': self.client_secret,
            'grant_type': constants.GRANT_TYPE_CREDENTIALS
        }
        timeout = self.timeout
        verbose = self.verbose
        max_retries = self.max_retries
        request = HTTPRequest(
            self.endpoint, max_retries=max_retries, verbose=verbose)
        result = request.post(
            '/oauth2/access_token', data=data, timeout=timeout)
        return models.ClientCredentialsToken(self, result)

    def get_api(self, access_token):
        """获取平台 API 调用客户端对象

        Args:
            access_token (str): 访问令牌；

        Returns:
            API: API 对象；
        """
        return API(
            access_token,
            endpoint=self.endpoint,
            max_retries=self.max_retries,
            timeout=self.timeout,
            verbose=self.verbose)


class API(object):
    """BGE 开放平台接口调用客户端

    Args:
        access_token (str): 访问令牌；
        endpoint (字符串, 非必填): 平台对外服务的访问域名，默认值为 pro-main；
        max_retries (数字, 非必填): 接口请求重试次数，默认值为 3；
        timeout (数字, 非必填): 接口请求默认超时间，默认值为 18；
        verbose (布尔, 非必填)：输出测试日志，默认值为 False；
    """

    alive = alive
    token_type = 'Bearer'

    def __init__(self, access_token, endpoint=None, max_retries=None,
                 timeout=None, verbose=False):
        self.access_token = access_token
        if endpoint is None:
            endpoint = constants.DEFAULT_ENDPOINT
        self.endpoint = endpoint
        if max_retries is not None:
            max_retries = int(max_retries)
        self.max_retries = max_retries
        if timeout is not None:
            timeout = int(timeout)
        self.timeout = timeout
        self.verbose = verbose
        self.logger = new_logger(self.__class__.__name__, verbose=verbose)

    def introspect(self):
        """验证当前使用的 access_token 有效性

        Returns:
            Model: Token 元数据；
        """
        timeout = self.timeout
        verbose = self.verbose
        max_retries = self.max_retries
        request = HTTPRequest(
            self.endpoint, max_retries=max_retries, verbose=verbose)
        request.set_authorization(self.token_type, self.access_token)
        result = request.get('/oauth2/introspect', params={
            'token': self.access_token
        }, timeout=timeout)
        return models.Model(result)

    def get_user(self, **params):
        """获取用户信息

        Returns:
            Model: 用户数据；
        """
        timeout = self.timeout
        verbose = self.verbose
        max_retries = self.max_retries
        request = HTTPRequest(
            self.endpoint, max_retries=max_retries, verbose=verbose)
        request.set_authorization(self.token_type, self.access_token)
        result = request.get('/user', params=params, timeout=timeout)
        return models.Model(result)

    def get_variants(self, biosample_id, rsids, **params):
        """根据rsid查询变异位点信息

        Args:
            biosample_id (str): 生物样品编号；
            rsids (str): 多个 rs 编号，逗号分割（必填）；如：rs1229984；
                         最多一次提供100个；

        Returns:
            list: 变异位点信息；
        """
        params.update({
            'rsids': rsids,
            'biosample_id': biosample_id.upper()
        })
        timeout = self.timeout
        verbose = self.verbose
        max_retries = self.max_retries
        request = HTTPRequest(
            self.endpoint, max_retries=max_retries, verbose=verbose)
        request.set_authorization(self.token_type, self.access_token)
        result = request.get('/variants', params=params, timeout=timeout)
        data = []
        for item in result:
            data.append(models.Model(item))
        return data

    def get_samples(self, biosample_ids=None, biosample_sites=None,
                    omics=None, project_ids=None, organisms=None,
                    data_availability=None, statuses=None,
                    next_page=None, limit=50, **kwargs):
        """获取样品列表

        授权码模式: 可通过本接口获取授权用户的样品；
        客户端模式: 可通过本接口获取客户端应用通过注册接口注册(或预先生成)的样品；

        Args:
            biosample_ids (str, 非必填): 生物样品 id，逗号分割多个；
            biosample_sites (str, 非必填): 采样部位，取值范围：1-15；
            omics (str, 非必填): 所属组学，取值范围：1-2；
            project_ids (str, 非必填): 所属项目，逗号分割多个；
            organisms (str, 非必填): 样品生物体，取值范围：1-3；
            data_availability (boolean, 非必填): 数据可用性；
            statuses (str, 非必填): 数据状态，详情见 BGE 开放平台文档；
            next_page (int, 非必填): 要获取的页码，默认值为 None；
            limit (int, 非必填): 每页返回数量，默认值为 50；

        Returns:
            list: 样品列表；
        """
        params = {}
        params.update(kwargs)
        page = 1
        if next_page is not None:
            page = next_page
        params.update({
            'biosample_ids': biosample_ids.upper(),
            'biosample_sites': biosample_sites,
            'omics': omics,
            'project_ids': project_ids,
            'organisms': organisms,
            'data_availability': data_availability,
            'statuses': statuses,
            'page': page,
            'limit': limit
        })
        timeout = self.timeout
        verbose = self.verbose
        max_retries = self.max_retries
        request = HTTPRequest(
            self.endpoint, max_retries=max_retries, verbose=verbose)
        request.set_authorization(self.token_type, self.access_token)
        result = request.get('/samples', params=params, timeout=timeout)
        result = models.Model(result)
        data = []
        for item in result['result']:
            data.append(models.Model(item))
        result['result'] = data
        return result

    def get_sample(self, biosample_id):
        """获取样品

        授权码模式: 可通过本接口获取授权用户的样品；
        客户端模式: 可通过本接口获取客户端应用通过注册接口注册(或预先生成)的样品；

        Args:
            biosample_id (str): 生物样品编号；

        Returns:
            Model: 样品数据；
        """
        url = '/samples/{}'.format(biosample_id.upper())
        timeout = self.timeout
        verbose = self.verbose
        max_retries = self.max_retries
        request = HTTPRequest(
            self.endpoint, max_retries=max_retries, verbose=verbose)
        request.set_authorization(self.token_type, self.access_token)
        result = request.get(url, timeout=timeout)
        return models.Model(result)

    def register_sample(self, external_sample_id, biosample_site,
                        project_id, **kwargs):
        """注册样品

        Args:
            external_sample_id (str): 外部生物样品id；
            biosample_site (int):  采样部位；
            project_id (str): 项目编号；
            **kwargs: 其他非必填数据，例：library_id="HWJBAYTGAA170328-18"；

        Returns:
            Model: 样品数据，包含生物样品编号；
        """
        data = {}
        data.update(kwargs)
        data.update({
            'external_sample_id': external_sample_id,
            'biosample_site': int(biosample_site),
            'project_id': project_id
        })
        timeout = self.timeout
        verbose = self.verbose
        max_retries = self.max_retries
        request = HTTPRequest(
            self.endpoint, max_retries=max_retries, verbose=verbose)
        request.set_authorization(self.token_type, self.access_token)
        result = request.post(
            '/samples/register', data=data, timeout=timeout)
        return models.Model(result)

    def improve_sample(self, biosample_id, **kwargs):
        """补充样品中未被赋值的信息

        已赋值数据无法变更，否则接口报错；

        Args:
            biosample_id (str): 生物样品编号；
            **kwargs: 需要赋值的字段和值；
        """
        if not kwargs:
            # 无更新
            return
        data = {}
        data.update(kwargs)
        data['biosample_id'] = biosample_id.upper()
        timeout = self.timeout
        verbose = self.verbose
        max_retries = self.max_retries
        request = HTTPRequest(
            self.endpoint, max_retries=max_retries, verbose=verbose)
        request.set_authorization(self.token_type, self.access_token)
        request.post(
            '/samples/improve', data=data, timeout=timeout)

    def get_taxon_abundance(self, biosample_id, taxon_ids=None,
                            next_page=None, limit=50, **params):
        """获取类群丰度

        Args:
            biosample_id (str): 生物样品编号；
            taxon_ids ([str], 非必填): BGE 物种编号，多个以逗号分割；
            next_page ([int], 非必填): 当前页码，默认值为 1，即首页；
            limit (int, 非必填): [description]. 默认值为 50；

        Returns:
            Model: 类群丰度数据详情；
        """
        params.update({
            'biosample_id': biosample_id.upper(),
            'taxon_id': taxon_ids,
            'limit': limit
        })
        page = 1
        if next_page is not None:
            page = next_page
        params['page'] = page
        timeout = self.timeout
        verbose = self.verbose
        max_retries = self.max_retries
        request = HTTPRequest(
            self.endpoint, max_retries=max_retries, verbose=verbose)
        request.set_authorization(self.token_type, self.access_token)
        result, pagination = request.get(
            '/microbiome/taxon_abundance', params=params, timeout=timeout)
        # TODO: upgrade in the future
        # 暂时特殊处理此接口，统一丰度数据的返回方式
        ret = models.Model({})
        ret['count'] = count = pagination['count']
        next_page = pagination['page'] + 1
        if count == 0:
            next_page = None
        ret['next_page'] = next_page
        ret['result'] = [models.Model(item) for item in result]
        return ret

    def get_func_abundance(self, biosample_id, catalog, ids=None, limit=50,
                           next_page=None, **params):
        """获取功能丰度

        Args:
            biosample_id (str): 生物样品编号；
            catalog (str): 目录标签，可选值为：go、ko、eggnog、pfam、kegg-pwy、
                           kegg-mdl、level4ec、metacyc-rxn、metacyc-pwy；
            ids (str, 非必填): BGE物种功能编号，多个值以逗号隔开；
            limit (int, 非必填): 一页返回数量，默认值为 50；
            next_page (str, 非必填): 下一页，用于获取下一页数据；

        Returns:
            Model: 功能丰度数据详情；
        """
        params.update({
            'biosample_id': biosample_id.upper(),
            'catalog': catalog,
            'ids': ids,
            'next_page': next_page,
            'limit': limit
        })
        timeout = self.timeout
        verbose = self.verbose
        max_retries = self.max_retries
        request = HTTPRequest(
            self.endpoint, max_retries=max_retries, verbose=verbose)
        request.set_authorization(self.token_type, self.access_token)
        result = request.get(
            '/microbiome/func_abundance', params=params, timeout=timeout)
        result = models.Model(result)
        result['result'] = [models.Model(item) for item in result['result']]
        return result

    def get_gene_abundance(self, biosample_id, catalog, data_type, ids=None,
                           limit=None, next_page=None, **params):
        """获取基因丰度

        Args:
            biosample_id (str): 生物样品编号；
            catalog (str): 分类标签，可选值：IGC_9.9M、UniRef90_HUMAnN2_0.11；
            data_type (str): 返回数据类型，可选值：list、file；
            ids (str, 非必填): BGE 物种 IGC 基因编号，多个值以逗号分割，
                                    如：igc_50,igc_51；
            limit (int, 非必填): 一页最大返回数量，默认 50，最大值为 1000；
            next_page (str, 非必填): 接口返回的下一页参数；

        Returns:
            Model: 基因丰度数据详情；
        """
        params.update({
            'biosample_id': biosample_id.upper(),
            'catalog': catalog,
            'data_type': data_type,
            'ids': ids,
            'next_page': next_page,
            'limit': limit
        })
        timeout = self.timeout
        verbose = self.verbose
        max_retries = self.max_retries
        request = HTTPRequest(
            self.endpoint, max_retries=max_retries, verbose=verbose)
        request.set_authorization(self.token_type, self.access_token)
        result = request.get(
            '/microbiome/gene_abundance', params=params, timeout=timeout)
        result = models.Model(result)
        result['result'] = [models.Model(item) for item in result['result']]
        return result

    def get_upload_token(self, **kwargs):
        """获取文件上传授权
        
        获取的授权仅包括当前目录（不含子目录）下的文件读、写权限；

        Returns:
            Model: 授权数据；
        """
        timeout = self.timeout
        verbose = self.verbose
        max_retries = self.max_retries
        request = HTTPRequest(
            self.endpoint, max_retries=max_retries, verbose=verbose)
        request.set_authorization(self.token_type, self.access_token)
        result = request.post('/sts/token', data=kwargs, timeout=timeout)
        return models.Model(result)

    def upload(self, filename, file_or_string, cmk_id=None):
        """上传文件

        Args:
            filename (str): 要上传到服务器的文件名；
            file_or_string (file-like-object or str): 文件内容或类文件对象；
            cmk_id (str)： 阿里云 KMS 服务用户主密钥 ID，加密上传时提供 CMK ID 即可；

        Returns:
            object_name: 文件的 OSS 对象名；
        """
        token = self.get_upload_token()
        return self._upload_file(token, filename, file_or_string, cmk_id=cmk_id)

    def batch_upload(self, files, cmk_id=None):
        """批量上传文件

        Args:
            files (FileItem object list): 要上传到服务器的文件列表；
            cmk_id (str)： 阿里云 KMS 服务用户主密钥 ID，加密上传时提供 CMK ID 即可；

        Returns:
            object_name: 文件的 OSS 对象名；
        """
        if not files:
            raise BGEError('files is required')
        object_names = []
        if isinstance(files, (list, tuple)):
            files = [files]
        token = self.get_upload_token()
        for file_obj in files:
            if not isinstance(file_obj, FileItem):
                continue
            filename = file_obj.filename
            file_or_string = file_obj.file_or_string
            sys.stdout.write('开始上传：{}\n'.format(filename))
            object_name = self._upload_file(
                token, filename, file_or_string, cmk_id=cmk_id)
            sys.stdout.write('\n\n')
            sys.stdout.flush()
            object_names.append(object_name)
        return object_names

    def upload_dir(self, dirpath, cmk_id=None):
        """上传目录下的文件(不递归上传子文件夹中文件)

        仅上传目录中的文件，软链接、符号链接、文件夹均不会上传至平台。

        Args:
            dirpath (str): 要上传到服务器的文件夹；
            cmk_id (str)： 阿里云 KMS 服务用户主密钥 ID，加密上传时提供 CMK ID 即可；

        Returns:
            object_names: 上传的文件 OSS 对象名列表；
        """
        token = self.get_upload_token()
        object_names = []
        for filename in os.listdir(dirpath):
            filepath = join(dirpath, filename)
            if isdir(filepath):
                continue
            sys.stdout.write('开始上传：{}\n'.format(filepath))
            with open(filepath, 'rb') as fp:
                object_name = self._upload_file(
                    token, filename, fp, cmk_id=cmk_id)
            sys.stdout.write('\n\n')
            sys.stdout.flush()
            object_names.append(object_name)
        return object_names

    def _upload_file(self, token, filename, file_or_string, cmk_id=None):
        token_meta = self.introspect()
        if token_meta['active'] == False:
            raise BGEError('access_token has expired')
        client_id = token_meta['client_id']
        credentials = token.credentials
        destination = token.destination
        bucket_name = token.bucket
        endpoint = token.endpoint
        access_key_id = credentials['access_key_id']
        access_key_secret = credentials['access_key_secret']
        security_token = credentials['security_token']
        auth = oss2.StsAuth(
            access_key_id, access_key_secret, security_token)
        if cmk_id is not None:
            region_id = token.region_id
            kms_provider = oss2.AliKMSProvider(
                access_key_id, access_key_secret, region_id, cmk_id)
            # NOTE 官方 oss2 处理 STS 加密上传存在 bug，等待其修复，此处做代码动态修改
            sts_token_credential = StsTokenCredential(
                access_key_id, access_key_secret, security_token)
            kms_provider.kms_client = AcsClient(
                region_id=region_id, credential=sts_token_credential)
            bucket = oss2.CryptoBucket(
                auth, endpoint, bucket_name, crypto_provider=kms_provider)
        else:
            bucket = oss2.Bucket(auth, endpoint, bucket_name)
        object_name = '%s/%s' % (destination, filename)
        bge_open_client_id_header = 'x-oss-meta-bge-open-client-id'
        custom_headers = { bge_open_client_id_header: client_id }
        bucket.put_object(
            object_name,
            file_or_string,
            headers=custom_headers,
            progress_callback=progress_callback)
        sys.stdout.write('')
        return object_name

    def get_download_url(self, object_name, region=None,
                         expiration_time=600, **kwargs):
        """获取阿里云OSS（对象存储）中的文件下载地址

        Args:
            object_name (str): OSS对象；
            region (str, 非必填): 区域（domestic、international），默认值为
                                 domestic；
            expiration_time (int, 非必填): 下载地址过期时间，默认值 600s；

        Returns:
            Model: 文件下载地址；
        """
        data = {}
        data.update(kwargs)
        data.update({
            'object_name': object_name,
            'region': region,
            'expiration_time': expiration_time
        })
        timeout = self.timeout
        verbose = self.verbose
        max_retries = self.max_retries
        request = HTTPRequest(
            self.endpoint, max_retries=max_retries, verbose=verbose)
        request.set_authorization(self.token_type, self.access_token)
        result = request.post('/oss/sign_url', data=data, timeout=timeout)
        return models.Model(result)

    def download(self, object_name, fp, region=None,
                 expiration_time=600, chunk_size=8192, **kwargs):
        """下载存储在阿里云OSS（对象存储）中的文件

        Args:
            object_name (str): OSS对象；
            fp(file like object): 可写的类文件对象；
            region (str, 非必填): 区域（domestic、international），默认值为
                                 domestic；
            chunk_size(int): 下载块大小
            expiration_time (int, 非必填): 下载地址过期时间，默认值 600s；
        """
        inst = self.get_download_url(
            object_name, region=None, expiration_time=expiration_time,
            **kwargs)
        size = 0
        prog_size = 61  # 单行输出的进度条固定为 80 个字符长度
        url = inst.url
        chunk_size = int(chunk_size)
        sys.stdout.write('\nStart downloading: %s' % object_name)
        sys.stdout.write('\n\n')
        timeout = self.timeout
        try:
            with requests.get(url, stream=True, timeout=timeout) as r:
                total = int(r.headers['content-length'])
                for chunk in r.iter_content(chunk_size):
                    size += len(chunk)
                    eq_size = int(size * prog_size / total)
                    equal_s = '=' * eq_size
                    blank_s = ' ' * (prog_size - eq_size)
                    progress = '>'.join((equal_s, blank_s))
                    percent = '%.2f%%' % float(size / total * 100)
                    sys.stdout.write(
                        '\r\t%s [%s]' % (percent.rjust(7), progress)
                    )
                    fp.write(chunk)
                flush_func = getattr(fp, 'flush', None)
                if flush_func:
                    flush_func()
        except requests.exceptions.HTTPError:
            raise BGEError('download request error')
        except Exception as e:
            raise BGEError(e)

    def aggregate_omics_data(self, data_element_id, time_dimension, start_time,
                             end_time=None, biosample_id=None,
                             interval=1, periods=100, **kwargs):
        """聚合组学数据（目前仅支持聚合数据流中符合平台设定 JSONPath 规则的数值型数据）

        Args:
            data_element_id (str, 必填): 数据元编号；
            time_dimension (str, 必填): 子聚合的时间维度，可选值：year, quarter,
                                        month, week, day, minute, second
            start_time (str, 必填): 数据流生成时间的起始时间；
            end_time (str, 非必填): 数据流生成时间的结束时间，为空时默认取当前时间；
            biosample_id (str, 非必填): 生物样品编号，客户端模式下为必填；
            interval (int, 非必填): 聚合时间维度间隔，默认:1
            periods (int, 非必填): 聚合时间维度返回数，默认:100，最大值：100
        
        Returns:
            Model: 返回的聚合数据
        """
        params = {}
        params.update(kwargs)
        params.update({
            'biosample_id': biosample_id.upper(),
            'data_element_id': data_element_id,
            'time_dimension': time_dimension,
            'start_time': start_time,
            'end_time': end_time,
            'interval': interval,
            'periods': periods
        })
        timeout = self.timeout
        verbose = self.verbose
        max_retries = self.max_retries
        request = HTTPRequest(
            self.endpoint, max_retries=max_retries, verbose=verbose)
        request.set_authorization(self.token_type, self.access_token)
        result = request.get(
            '/omics_data/aggregate', params=params, timeout=timeout)
        return models.Model(result)

    def get_range_stream(self, data_element_id, biosample_id=None,
                         start_time=None, end_time=None,
                         sort_direction=None, limit=100, next_page=None,
                         **kwargs):
        """返回查询数据流
        
        
        Args:
            biosample_id (str, 非必填): 生物样品编号，客户端模式下为必填；
            start_time (str, 非必填): 数据流生成时间的起始时间
            end_time (str, 非必填): 数据流生成时间的结束时间
            sort_direction (str, 非必填): 排序方式，默认：desc
            limit (int, 非必填): 每页返回数量，默认:100
            next_page (str, 非必填): 下一页参数
        
        Returns:
            Model: 返回的数据流数据
        """
        params = {}
        params.update(kwargs)
        params.update({
            'biosample_id': biosample_id.upper(),
            'data_element_id': data_element_id,
            'start_time': start_time,
            'end_time': end_time,
            'sort_direction': sort_direction,
            'limit': limit,
            'next_page': next_page
        })
        timeout = self.timeout
        verbose = self.verbose
        max_retries = self.max_retries
        request = HTTPRequest(
            self.endpoint, max_retries=max_retries, verbose=verbose)
        request.set_authorization(self.token_type, self.access_token)
        result = request.get(
            '/stream/range', params=params, timeout=timeout)
        return models.Model(result)

    def get_data_items(self, namespace, biosample_id,
                       collection_id=None, data_element_ids=None,
                       next_page=None, **kwargs):
        """根据数据元编号或数据集编号获取数据项

        Args:
            namespace（str）：命名空间
            biosample_id (str): 生物样品编号；
            collection_id（str，非必填）：数据集编号，与 data_element_ids 互斥；
            data_element_ids (str，非必填): 多个数据元编号，逗号分割（必填）；
                                           最多一次提供100个；
            next_page (int，非必填): 下一页；

        Returns:
            Model: 返回的数据项数据；
        """
        params = {}
        params.update(kwargs)
        params.update({
            'biosample_id': biosample_id.upper(),
            'data_element_ids': data_element_ids,
            'collection_id': collection_id,
            'namespace': namespace,
            'next_page': next_page
        })
        timeout = self.timeout
        verbose = self.verbose
        max_retries = self.max_retries
        request = HTTPRequest(
            self.endpoint, max_retries=max_retries, verbose=verbose)
        request.set_authorization(self.token_type, self.access_token)
        result = request.get(
            '/data_item/batch_retrieve', params=params, timeout=timeout)
        return models.Model(result)

    def invoke_model(self, model_id, **kwargs):
        """模型调用

        Args:
            model_id (str): 模型编号；
            **kwargs: 模型相关参数，由模型定义的参数决定；

        Returns:
            Model: 任务id，时间和状态；
        """
        params = {}
        params.update(kwargs)
        timeout = self.timeout
        verbose = self.verbose
        max_retries = self.max_retries
        request = HTTPRequest(
            self.endpoint, max_retries=max_retries, verbose=verbose)
        request.set_authorization(self.token_type, self.access_token)
        model_url = '/model/{}'.format(model_id)
        result = request.get(model_url, params=params, timeout=timeout)
        return models.Model(result)

    def invoke_draft_model(self, model_id, **kwargs):
        """调用灰度部署版本模型

        Args:
            model_id (str): 模型编号；
            **kwargs: 模型相关参数，由模型定义的参数决定；

        Returns:
            Model: 任务id，时间和状态；
        """
        params = {}
        params.update(kwargs)
        timeout = self.timeout
        verbose = self.verbose
        max_retries = self.max_retries
        request = HTTPRequest(
            self.endpoint, max_retries=max_retries, verbose=verbose)
        request.set_authorization(self.token_type, self.access_token)
        model_url = '/model/{}/draft'.format(model_id)
        result = request.get(model_url, params=params, timeout=timeout)
        return models.Model(result)

    def deploy_model(self, model_id, object_name=None, runtime=None,
                     memory_size=None, timeout=None, **kwargs):
        """部署模型

        Args:
            model_id (str): 模型编号。
            object_name (str): 模型源代码文件阿里云 OSS 对象路径。
            runtime (str, optional): 运行环境版本。默认值：python3。
            memory_size (int, optional): 内存占用量，单位：MB。默认：128MB。
            timeout (int, 非必填): 模型运行超时时间，单位：秒。默认：900。

        Returns:
            Model: 任务id，时间和状态；
        """
        data = {}
        data.update(kwargs)
        data.update({
            'model_id': model_id,
            'runtime': runtime,
            'memory_size': memory_size,
            'timeout': timeout,
            'object_name': object_name
        })
        timeout = self.timeout
        verbose = self.verbose
        max_retries = self.max_retries
        request = HTTPRequest(
            self.endpoint, max_retries=max_retries, verbose=verbose)
        request.set_authorization(self.token_type, self.access_token)
        result = request.post(
            '/model/deploy', data=data, timeout=timeout)
        return models.Model(result)

    def publish_model(self, model_id, message, **kwargs):
        """发布模型稳定版

        Args:
            model_id (str): 模型编号。
            message (str): 本次发布模型的描述信息。

        Returns:
            [int]: 稳定版模型发布成功后返回的版本号
        """

        data = {}
        data.update(kwargs)
        data.update({
            'model_id': model_id,
            'message': message
        })
        timeout = self.timeout
        verbose = self.verbose
        max_retries = self.max_retries
        request = HTTPRequest(
            self.endpoint, max_retries=max_retries, verbose=verbose)
        request.set_authorization(self.token_type, self.access_token)
        result = request.post(
            '/model/publish', data=data, timeout=timeout)
        return models.Model(result)

    def rollback_model(self, model_id, version, **kwargs):
        """回滚模型

        Args:
            model_id (str): 模型编号。
            version (int): 模型版本号。

        Returns:
            [str]: 模型回滚成功后返回的新的版本号
        """
        data = {}
        data.update(kwargs)
        data.update({
            'model_id': model_id,
            'version': version
        })
        timeout = self.timeout
        verbose = self.verbose
        max_retries = self.max_retries
        request = HTTPRequest(
            self.endpoint, max_retries=max_retries, verbose=verbose)
        request.set_authorization(self.token_type, self.access_token)
        result = request.post(
            '/model/rollback', data=data, timeout=timeout)
        return models.Model(result)

    def model_versions(self, model_id, limit=10, next_page=None, **kwargs):
        """模型历史版本列表

        Args:
            model_id (str): 模型编号。
            limit (int): 每页返回数量，默认值为 10。
            next_page (int): 下一页；

        Returns:
            [list]: 模型历史版本列表
        """
        params = {}
        params.update(kwargs)
        params['limit'] = limit
        params['next_page'] = next_page
        timeout = self.timeout
        verbose = self.verbose
        max_retries = self.max_retries
        request = HTTPRequest(
            self.endpoint, max_retries=max_retries, verbose=verbose)
        request.set_authorization(self.token_type, self.access_token)
        model_url = '/model/{}/versions'.format(model_id)
        result = request.get(model_url, params=params, timeout=timeout)
        return models.Model(result)

    def upload_model_expfs(self, model_id, expfs, **kwargs):
        """上传模型扩展文件集

        Args:
            model_id (str): 模型编号。
            expfs (path): 模型扩展文件集。

        Returns:
            Model: 任务id、时间、状态和返回值；
        """
        if isinstance(expfs, text_type):
            filename = split(expfs)[1]
        else:
            filename = expfs.name
        def upload_callback(monitor):
            total_bytes = monitor.len
            bytes_consumed = monitor.bytes_read
            sys.stdout.write(
                '\r文件大小：{}, 上传进度: {}%，已上传 {}'.format(
                    human_byte(total_bytes, 2),
                    '%.2f' % ((bytes_consumed / total_bytes) * 100),
                    human_byte(bytes_consumed, 2)
                )
            )
            sys.stdout.flush()
        e = encoder.MultipartEncoder(
            fields={
                'model_id': model_id,
                'expfs': (filename, open(expfs, 'rb'), 'application/zip')
            }
        )
        m = encoder.MultipartEncoderMonitor(e, upload_callback)
        timeout = self.timeout
        verbose = self.verbose
        max_retries = self.max_retries
        request = HTTPRequest(
            self.endpoint, max_retries=max_retries, verbose=verbose)
        request.set_authorization(self.token_type, self.access_token)
        model_url = '/model/expfs/upload'
        result = request.post(model_url, data=m, timeout=timeout, headers={
            'Content-Type': m.content_type
        })
        sys.stdout.write('')
        return models.Model(result)

    def task(self, task_id):
        """获取任务结果

        Args:
            task_id (str): 任务编号。

        Returns:
            Model: 任务id、时间、状态和返回值；
        """
        timeout = self.timeout
        verbose = self.verbose
        max_retries = self.max_retries
        request = HTTPRequest(
            self.endpoint, max_retries=max_retries, verbose=verbose)
        request.set_authorization(self.token_type, self.access_token)
        model_url = '/task/{}'.format(task_id)
        result = request.get(model_url, timeout=timeout)
        return models.Model(result)

    def upload_model_doc(self, doc_tab, model_id, doc_content):
        """上传模型文档
        
        Args:
            doc_tab (str): 文档所在的 tab。
            model_id (str): 模型编号。
            doc_content (list): 文档内容
        Returns:
            Model_id: 模型编号；
            version: 文档版本号。
        """
        data = {}
        data['doc_tab'] = doc_tab
        data['model_id'] = model_id
        data['doc_content'] = doc_content
        doc = json.dumps(data)
        timeout = self.timeout
        verbose = self.verbose
        max_retries = self.max_retries
        request = HTTPRequest(
            self.endpoint, max_retries=max_retries, verbose=verbose)
        request.set_authorization(self.token_type, self.access_token)
        result = request.post(
            '/model/doc_upload', data=doc, timeout=timeout)
        return models.Model(result)



import argparse
import docker
import fnmatch
import json
import os
import qprompt
import re
import stat
import sys
import tempfile
import zipfile

from datetime import datetime
from posixpath import join, exists, isdir, relpath
from six.moves import input
from time import sleep
from uuid import uuid4

from bgesdk.client import API
from bgesdk.error import APIError
from bgesdk.management import constants
from bgesdk.management.command import BaseCommand
from bgesdk.management.utils import (
    get_active_project, config_get, get_home, read_config,
    generate_container_name, get_config_parser, confirm,
    output, get_sys_user, SYS_STR
)
from bgesdk.utils import human_byte
from bgesdk.version import __version__


DEFAULT_OAUTH2_SECTION = constants.DEFAULT_OAUTH2_SECTION
DEFAULT_TOKEN_SECTION = constants.DEFAULT_TOKEN_SECTION
DEFAULT_MODEL_SECTION = constants.DEFAULT_MODEL_SECTION
DEFAULT_MODEL_TIMEOUT = constants.DEFAULT_MODEL_TIMEOUT
TEST_SERVER_ENDPOINT = constants.TEST_SERVER_ENDPOINT
TEST_SERVER_PORT = constants.TEST_SERVER_PORT

model_id_match = re.compile(r'^[a-zA-Z0-9]+$').match

MAIN_PY = '''\
import json
import bgesdk
import logging

def handler(event, context):
    logging.debug(event)
    event = json.loads(event)
    biosample_id = event['biosample_id']
    access_token = event['access_token']
    params = event['params']
    return json.dumps({
        "model_code": 0,
        "model_msg": "success",
        "model_data": {
            'biosample_id': biosample_id,
            'params': params,
            'sdk': {
                'version': bgesdk.__version__
            }
        }
    })
'''

MODEL_CONFIG_TEMPLATE = '''\
[Model]
model_id =
memory_size = 128
timeout = 900
runtime = python3
'''

BGEIGNORE_TEMPLATE = '''\
# Unix filename patterns
#
#   *    matches everything
#   ?    matches any single character
#   seq  matches any character in seq
#   !seq matches any character not in seq


# Ignore folder .bge/

.bge/*

'''

RUNTIMES = {
    'python2.7': 'teambge/model-python2.7:latest',
    'python3': 'teambge/model-python3.6:latest'
}
WORKDIR = '/code'

RUNTIME_CHOICES = ['python3', 'python2.7']
MEMORY_SIZE_CHOICES = [
    128, 192, 256, 320, 384, 448, 512, 576, 640, 704, 768, 832,
    896, 960, 1024, 1088, 1152, 1216, 1280, 1344, 1408, 1472, 1536,
    1600, 1664, 1728, 1792, 1856, 1920, 1984, 2048
]
MIN_TIMEOUT = 1
MAX_TIMEOUT = 900

TOTAL_PROGRESS = {
    'FAILURE': '任务执行失败',
    'PENDING': '任务等待中',
    'RECEIVED': '任务已接收',
    'RETRY': '任务将被重试',
    'REVOKED': '任务取消',
    'STARTED': '任务已开始',
    'SUCCESS': '任务执行成功'
}


class Command(BaseCommand):

    order = 5
    help='模型初始化脚手架、配置、部署等相关命令。'

    def add_arguments(self, parser):
        home = get_home()
        model_subparsers = parser.add_subparsers(
            dest='subcommand',
            help='可选子命令。'
        )

        # 初始化脚手架命令
        init_p = model_subparsers.add_parser(
            'init',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            help='初始化一个新的模型开发脚手架项目。'
        )
        init_p.add_argument(
            'scaffold_name',
            type=str,
            help='脚手架名字。'
        )
        init_p.add_argument(
            '--home',
            type=str,
            default=home,
            help='脚手架项目生成的父级目录，默认为当前目录。'
        )
        init_p.set_defaults(method=self.init_scaffold, parser=init_p)

        # 模型配置
        config_p = model_subparsers.add_parser(
            'config',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            help='配置模型。'
        )
        config_p.add_argument(
            '-s',
            '--show',
            default=False,
            action='store_true',
            help='打印显示当前模型脚手架的配置。'
        )
        config_p.set_defaults(method=self.config_model, parser=config_p)

        # 安装 Python 依赖包
        install_p = model_subparsers.add_parser(
            'install',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            help='部署灰度版本模型。（开发中，仅支持 python2.7/3.6 的 docker 环境）'
        )
        install_p.add_argument(
            'package_name',
            nargs="+",
            type=str,
            help='要安装的 Python 软件包，如 numpy 或 numpy==v1.19.5'
        )
        install_p.add_argument(
            '-r',
            '--requirements',
            nargs=1,
            type=str,
            help='要安装的 Python 软件包依赖文件'
        )
        install_p.add_argument(
            '-U',
            '--upgrade',
            default=False,
            action='store_true',
            help='是否升级依赖包，同 pip install -U'
        )
        install_p.add_argument(
            '--force-reinstall',
            default=False,
            action='store_true',
            help='是否强制重新安装依赖包，同 pip install --force-reinstall'
        )
        install_p.set_defaults(method=self.install_deps, parser=install_p)

        start_p = model_subparsers.add_parser(
            'start',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            help='启动本地 HTTP 测试服务器'
        )
        start_p.set_defaults(method=self.start_model, parser=start_p)

        expfs_p = model_subparsers.add_parser(
            'expfs',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            help='扩展模型文件集'
        )
        expfs_p.set_defaults(method=self.upload_model_expfs, parser=expfs_p)

        # 部署子命令
        deploy_p = model_subparsers.add_parser(
            'deploy',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            help='部署灰度版本模型。'
        )
        deploy_p.add_argument(
            '-i',
            '--ignore-source',
            default=False,
            action="store_true",
            help='部署时不包含模型源代码'
        )
        deploy_p.set_defaults(method=self.deploy_model, parser=deploy_p)
        publish_p = model_subparsers.add_parser(
            'publish',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            help='发布最新部署的灰度版本模型为稳定版。'
        )
        publish_p.add_argument(
            '-m',
            '--message',
            type=str,
            help='模型发布说明。'
        )
        publish_p.set_defaults(method=self.publish_model, parser=publish_p)

        # 运行模型子命令
        run_p = model_subparsers.add_parser(
            'run',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            help='调用线上稳定版模型。'
        )
        group = run_p.add_mutually_exclusive_group()
        group.add_argument(
            '-d',
            '--draft',
            default=False,
            action="store_true",
            help='调用最新灰度版本模型。'
        )
        group.add_argument(
            '-t',
            '--test',
            default=False,
            action="store_true",
            help='调用本地启动的测试服务器运行模型。'
        )
        group = run_p.add_mutually_exclusive_group()
        group.add_argument(
            '-a',
            '--args',
            nargs=2,
            action='append',
            help='参数对，示例：--args f1 v1 --args f2 v2'
        )
        group.add_argument(
            '-f',
            '--file',
            help='JSON 格式参数文件，内容示例：{ "f1": "v1", "f2": "v2" }。'
        )
        run_p.set_defaults(method=self.run_model, parser=run_p)

        # 回滚子命令
        rollback_p = model_subparsers.add_parser(
            'rollback',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            help='模型回滚至某个稳定版本。'
        )
        rollback_p.add_argument(
            '-v',
            '--version',
            type=str,
            help='要回滚的模型版本。'
        )
        rollback_p.set_defaults(method=self.rollback_model, parser=rollback_p)

        # 回滚版本列表子命令
        versions_p = model_subparsers.add_parser(
            'versions',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            help='模型发布版本列表。（开发中）'
        )
        versions_p.add_argument(
            '-l',
            '--limit',
            default=5,
            type=int,
            help='每页返回数量。'
        )
        versions_p.add_argument(
            '-p',
            '--next_page',
            type=int,
            help='下一页。'
        )
        versions_p.set_defaults(method=self.model_versions, parser=versions_p)

    def handler(self, args):
        """打印 subparser 帮助信息"""
        parser = args.parser
        parser.print_help(sys.stderr)

    def init_scaffold(self, args):
        scaffold_name = args.scaffold_name
        home = args.home
        if home is None:
            home = get_home()
        scaffold_dir = join(home, scaffold_name)
        if exists(scaffold_dir):
            output('错误！{} 已存在 '.format(scaffold_dir))
            sys.exit(1)
        if not exists(home):
            output('错误！无法找到 home 目录 {}。'.format(home))
            sys.exit(1)
        model_id, runtime, memory_size, timeout = self._config_model()
        os.makedirs(scaffold_dir)
        bge_dir = join(scaffold_dir, '.bge')
        lib_dir = join(scaffold_dir, 'lib')
        for dir_ in [scaffold_dir, bge_dir, lib_dir]:
            sys.stdout.write('创建 {} ... '.format(dir_))
            sys.stdout.flush()
            if not exists(dir_):
                os.makedirs(dir_)
                output('完成')
            elif not isdir(dir_):
                output('失败')
                output('失败！{} 存在但不是目录。'.format(dir_))
                sys.exit(1)
            else:
                output('已存在')
        model_config_path = join(scaffold_dir, 'model.ini')
        output('创建 {} ... '.format(model_config_path))
        if not exists(model_config_path):
            open(model_config_path, 'w').write(MODEL_CONFIG_TEMPLATE)
        ignore_path = join(scaffold_dir, '.bgeignore')
        output('创建 {} ... '.format(ignore_path))
        if not exists(ignore_path):
            open(ignore_path, 'w').write(BGEIGNORE_TEMPLATE)
        script_name = 'main.py'
        script_path = join(scaffold_dir, script_name)
        with open(script_path, 'wb') as file_out:
            file_out.write(MAIN_PY.encode())
        st = os.stat(script_path)
        os.chmod(script_path, st.st_mode | stat.S_IEXEC)
        self._save_model_config(
            model_id, runtime, memory_size, timeout, home=scaffold_dir)
        if confirm('是否安装 bge-python-sdk？'):
            os.chdir(scaffold_dir)
            self._install_sdk()
        output('成功创建模型项目脚手架')

    def config_model(self, args):
        if args.show:
            return self._read_model_project(args)
        config_path = self.get_model_config_path()
        config_parser = get_config_parser(config_path)
        model_id, runtime, memory_size, timeout = self._config_model(config_parser)
        self._save_model_config(model_id, runtime, memory_size, timeout)

    def _config_model(self, config=None):
        output('开始配置解读模型...')
        default_model_id = ''
        default_runtime = 'python3'
        default_memory_size = 128
        default_timeout = 900
        if config is not None:
            section_name = DEFAULT_MODEL_SECTION
            default_model_id = config_get(config.get, section_name, 'model_id')
            default_runtime = config_get(config.get, section_name, 'runtime')
            default_memory_size = config_get(
                config.getint, section_name, 'memory_size')
            default_timeout = config_get(
                config.getint, section_name, 'timeout')
        # model_id
        prompt = '？请输入解读模型编号 [{}]：'.format(default_model_id)
        while True:
            model_id = input(prompt)
            if not model_id_match(model_id):
                output('模型编号只能包含数字、大小写字母')
                continue
            if model_id:
                break
            if not default_model_id:
                output('模型编号未设置...')
                continue
            model_id = default_model_id
            break
        # runtime
        try:
            index = RUNTIME_CHOICES.index(default_runtime)
        except IndexError:
            index = 0
        prompt = '？请选择运行环境 [{}]'.format(default_runtime)
        input_prompt = "请选择"
        runtime = qprompt.enum_menu(RUNTIME_CHOICES).show(header=prompt,
                                                          dft=index + 1,
                                                          returns="desc",
                                                          msg=input_prompt)
        # memory_size
        try:
            index = MEMORY_SIZE_CHOICES.index(default_memory_size)
        except IndexError:
            index = 0
        prompt = '？请选择内存占用（MB） [{}]'.format(default_memory_size)
        input_prompt = "请选择"
        memory_size = qprompt.enum_menu(MEMORY_SIZE_CHOICES).show(header=prompt,
                                                                  dft=index + 1,
                                                                  returns="desc",
                                                                  msg=input_prompt)
        # timeout
        prompt = '请输入模型运行超时时间，类型为整数'
        while True:
            timeout = qprompt.ask_int(msg=prompt, dft=default_timeout)
            if (MIN_TIMEOUT and timeout < MIN_TIMEOUT) \
                    or (MAX_TIMEOUT and timeout > MAX_TIMEOUT):
                output('超出范围 [{},{}]，请重新输入'.format(
                    MIN_TIMEOUT or '', MAX_TIMEOUT or ''))
                continue
            break
        if not timeout:
            timeout = default_timeout
        return model_id, runtime, memory_size, timeout

    def _save_model_config(self, model_id, runtime, memory_size, timeout, home=None):
        # save
        config_path = self.get_model_config_path(home=home)
        config = get_config_parser(config_path)
        section_name = DEFAULT_MODEL_SECTION
        if section_name not in config.sections():
            config.add_section(section_name)
        output('开始保存解读模型配置...')
        output('配置文件 {}'.format(config_path))
        config.set(section_name, 'model_id', model_id)
        config.set(section_name, 'runtime', runtime)
        config.set(section_name, 'memory_size', str(memory_size))
        config.set(section_name, 'timeout', str(timeout))
        with open(config_path, 'w') as config_file:
            config.write(config_file)
        output('')
        output('解读模型配置已保存至：{}'.format(config_path))

    def _install_sdk(self):
        home = get_home()
        config_path = self.get_model_config_path()
        config = get_config_parser(config_path)
        section_name = DEFAULT_MODEL_SECTION
        model_id = config_get(config.get, section_name, 'model_id')
        runtime = config_get(config.get, section_name, 'runtime')
        image_name = RUNTIMES[runtime]
        container_name = generate_container_name(model_id)
        deps = []
        command = 'whereis docker' if SYS_STR == 'windows' else 'which docker'
        with os.popen(command) as f:
            docker_path = f.read()
        if not docker_path:
            output(
                '请先安装 docker，参考 https://docs.docker.com/engine/install/')
            sys.exit(1)
        try:
            client = docker.from_env()
        except docker.errors.DockerException:
            output('请确认 docker 服务是否已开启。')
            output('启动命令：/etc/init.d/docker start')
            sys.exit(1)
        command = ('pip install --no-deps bge-python-sdk pimento '
                   'requests_toolbelt -t /code/lib')
        try:
            client.images.get(image_name)
        except docker.errors.NotFound:
            output('本地 docker 镜像 {} 不存在，开始拉取...'.format(image_name))
            return_code = os.system('docker pull {}'.format(image_name))
            if return_code != 0:
                output('镜像拉取失败，请重试')
                sys.exit(1)
            output('镜像拉取成功')
        output('开始安装模型依赖包...')
        command = ('sh -c "{}"').format(command)
        user = get_sys_user()
        try:
            container = client.containers.get(container_name)
            if container.status != 'exited':
                try:
                    container.remove(force=True)
                except:
                    pass
        except docker.errors.NotFound:
            pass
        container = client.containers.run(
            image_name,
            command=command,
            name=container_name,
            volumes={home: { 'bind': WORKDIR, 'mode': 'rw' }},
            stop_signal='SIGINT',
            user=user,
            detach=True,
            auto_remove=True)
        try:
            logs = container.logs(stream=True)
            for log in logs:
                sys.stdout.write(log.decode('utf-8'))
                sys.stdout.flush()
        finally:
            if container.status != 'exited':
                try:
                    container.remove(force=True)
                except:
                    pass
        output('安装完成')

    def upload_model_expfs(self, args):
        home = get_home()
        project = get_active_project()
        config_path = self.get_model_config_path()
        config = get_config_parser(config_path)
        section_name = DEFAULT_MODEL_SECTION
        model_id = config_get(config.get, section_name, 'model_id')
        oauth2_section = DEFAULT_OAUTH2_SECTION
        token_section = DEFAULT_TOKEN_SECTION
        config = read_config(project)
        access_token = config_get(config.get, token_section, 'access_token')
        endpoint = config_get(config.get, oauth2_section, 'endpoint')
        api = API(
            access_token, endpoint=endpoint, timeout=DEFAULT_MODEL_TIMEOUT)
        try:
            result = api.upload_model_expfs(model_id, 'uploadtest.zip')
        except APIError as e:
            output('上传模型扩展集失败：{}'.format(e))
            sys.exit(1)
        task_id = result.task_id
        output('上传模型扩展集任务：{}'.format(task_id))
        task_path = join(home, '.bge', 'expfs.task_id')
        with open(task_path, 'w') as f:
            f.write(task_id)
        output('上传模型扩展集任务返回结果：')
        seconds = 0
        while True:
            result = api.task(task_id)
            progress = result.progress
            if progress not in TOTAL_PROGRESS:
                output('已等待 {}s，任务状态异常：{}'.format(seconds, progress))
                try:
                    os.unlink(task_path)
                except (IOError, OSError):
                    pass
                sys.exit(1)
            output('已等待 {}s，{}'.format(seconds, TOTAL_PROGRESS[progress]))
            if progress in ('SUCCESS', 'FAILURE', 'REVOKED'):
                break
            sleep(1)
            seconds += 1
        try:
            os.unlink(task_path)
        except (IOError, OSError):
            pass
        if 'SUCCESS' == progress:
            output('模型 {} 上传模型扩展集成功。'.format(model_id))
        elif 'FAILURE' == progress:
            output('模型 {} 上传模型扩展集失败。'.format(model_id))
        elif 'REVOKED' == progress:
            output('模型 {} 上传模型扩展集任务已被撤销。'.format(model_id))

    def deploy_model(self, args):
        """部署模型"""
        ignore_source = args.ignore_source
        home = get_home()
        project = get_active_project()
        config_path = self.get_model_config_path()
        config = get_config_parser(config_path)
        section_name = DEFAULT_MODEL_SECTION
        model_id = config_get(config.get, section_name, 'model_id')
        timestr = datetime.now().strftime('%Y%m%d%H%M%S%f')
        randstr = uuid4().hex
        zip_filename = '{}.{}.{}.zip'.format(model_id, timestr, randstr)
        params = {}
        params['runtime'] = config_get(config.get, section_name, 'runtime')
        params['memory_size'] = config_get(
            config.getint, section_name, 'memory_size')
        params['timeout'] = config_get(config.getint, section_name, 'timeout')
        oauth2_section = DEFAULT_OAUTH2_SECTION
        token_section = DEFAULT_TOKEN_SECTION
        config = read_config(project)
        access_token = config_get(config.get, token_section, 'access_token')
        endpoint = config_get(config.get, oauth2_section, 'endpoint')
        api = API(
            access_token, endpoint=endpoint, timeout=DEFAULT_MODEL_TIMEOUT)
        object_name = None
        if not ignore_source:
            ignore_path = join(home, '.bgeignore')
            if not exists(ignore_path):
                output('未发现 .bgeignore 文件，初始化 {} ... '.format(ignore_path))
                open(ignore_path, 'w').write(BGEIGNORE_TEMPLATE)
            output('开始打包模型源码...')
            tmp = tempfile.NamedTemporaryFile(suffix='.zip', delete=False)
            with zipfile.ZipFile(tmp.name, 'w', zipfile.ZIP_DEFLATED) as zf:
                self._zip_codedir(home, zf)
            tmp.flush()
            tmp.seek(0, 2)
            size = tmp.tell()
            tmp.seek(0)
            output('打包成功')
            human_size = human_byte(size)
            if size > 100 * 1024 * 1024:
                output('打包后 zip 文件为 {}，最大限制 100MB'.format(human_size))
                exit(1)
            output('打包 zip 文件大小为：{}'.format(human_size))
            output('开始上传模型源码...')
            try:
                object_name = api.upload(zip_filename, tmp)
            except APIError as e:
                output('上传模型源码失败：{}'.format(e))
                sys.exit(1)
            output('\n上传成功')
            tmp.close()
            os.remove(tmp.name)
        output('模型部署中...')
        try:
            result = api.deploy_model(
                model_id, object_name=object_name, **params)
        except APIError as e:
            output('模型部署失败：{}'.format(e))
            sys.exit(1)
        task_id = result.task_id
        output('模型部署任务：{}'.format(task_id))
        task_path = join(home, '.bge', 'task_id')
        with open(task_path, 'w') as f:
            f.write(task_id)
        output('模型部署任务返回结果：')
        seconds = 0
        while True:
            result = api.task(task_id)
            progress = result.progress
            if progress not in TOTAL_PROGRESS:
                output('已等待 {}s，任务状态异常：{}'.format(seconds, progress))
                try:
                    os.unlink(task_path)
                except (IOError, OSError):
                    pass
                sys.exit(1)
            output('已等待 {}s，{}'.format(seconds, TOTAL_PROGRESS[progress]))
            if progress in ('SUCCESS', 'FAILURE', 'REVOKED'):
                break
            sleep(1)
            seconds += 1
        try:
            os.unlink(task_path)
        except (IOError, OSError):
            pass
        if 'SUCCESS' == progress:
            output('模型 {} 灰度部署成功。'.format(model_id))
        elif 'FAILURE' == progress:
            output('模型 {} 灰度部署失败。任务结果：{}'.format(model_id, result))
        elif 'REVOKED' == progress:
            output('模型 {} 灰度部署任务已被撤销。'.format(model_id))

    def _zip_codedir(self, home, ziph):
        home = home.rstrip('/')
        total_files = {}
        zip_relpaths = set()
        for root, dirs, files in os.walk(home):
            for file in files:
                filepath = join(root, file)
                zip_relpath = relpath(filepath, home)
                zip_relpaths.add(zip_relpath)
                total_files[zip_relpath] = filepath
        src_zip_relpaths = zip_relpaths.copy()
        ignore_patterns = self.get_ignore_patterns(home)
        for ignore_pattern in ignore_patterns:
            if ignore_pattern.startswith('!'):
                ignore_pattern = ignore_pattern[1:]
                zip_relpaths.update(
                    fnmatch.filter(src_zip_relpaths, ignore_pattern))
            else:
                zip_relpaths -= set(
                    fnmatch.filter(zip_relpaths, ignore_pattern))
        for zip_relpath in zip_relpaths:
            filepath = total_files[zip_relpath]
            ziph.write(filepath, zip_relpath)

    def get_ignore_patterns(self, home):
        ignore_patterns = []
        home = home.rstrip('/')
        ignore_path = join(home, '.bgeignore')
        with open(ignore_path) as fp:
            for _ in fp.readlines():
                _ = _.strip()
                while _.startswith('/'):
                    _ = _[1:]
                if _.startswith('#') or not _:
                    continue
                ignore_patterns.append(_)
        return ignore_patterns

    def publish_model(self, args):
        """发布模型到稳定版"""
        message = args.message
        home = get_home()
        project = get_active_project()
        config_path = self.get_model_config_path()
        config = get_config_parser(config_path)
        section_name = DEFAULT_MODEL_SECTION
        model_id = config_get(config.get, section_name, 'model_id')
        oauth2_section = DEFAULT_OAUTH2_SECTION
        token_section = DEFAULT_TOKEN_SECTION
        config = read_config(project)
        access_token = config_get(config.get, token_section, 'access_token')
        endpoint = config_get(config.get, oauth2_section, 'endpoint')
        api = API(
            access_token, endpoint=endpoint, timeout=DEFAULT_MODEL_TIMEOUT)
        try:
            result = api.publish_model(model_id, message)
        except APIError as e:
            output('模型发布失败：{}'.format(e))
            sys.exit(1)
        output('模型 {} 稳定版已成功发布。\n版本号：{}。'.format(
            model_id, result['version']))

    def rollback_model(self, args):
        version = args.version
        home = get_home()
        project = get_active_project()
        config_path = self.get_model_config_path()
        config = get_config_parser(config_path)
        section_name = DEFAULT_MODEL_SECTION
        model_id = config_get(config.get, section_name, 'model_id')
        oauth2_section = DEFAULT_OAUTH2_SECTION
        token_section = DEFAULT_TOKEN_SECTION
        config = read_config(project)
        access_token = config_get(config.get, token_section, 'access_token')
        endpoint = config_get(config.get, oauth2_section, 'endpoint')
        api = API(
            access_token, endpoint=endpoint, timeout=DEFAULT_MODEL_TIMEOUT)
        try:
            result = api.rollback_model(model_id, version)
        except APIError as e:
            output('模型回滚失败：{}'.format(e))
            sys.exit(1)
        task_id = result.task_id
        output('模型回滚任务：{}'.format(task_id))
        task_path = join(home, '.bge', 'task_id')
        with open(task_path, 'w') as f:
            f.write(task_id)
        seconds = 0
        while True:
            result = api.task(task_id)
            progress = result.progress
            if progress not in TOTAL_PROGRESS:
                output('已等待 {}s，任务状态异常：{}'.format(seconds, progress))
                try:
                    os.unlink(task_path)
                except (IOError, OSError):
                    pass
                sys.exit(1)
            output('已等待 {}s，{}'.format(seconds, TOTAL_PROGRESS[progress]))
            if progress in ('SUCCESS', 'FAILURE', 'REVOKED'):
                break
            sleep(1)
            seconds += 1
        try:
            os.unlink(task_path)
        except (IOError, OSError):
            pass
        if 'SUCCESS' == progress:
            output('模型 {} 灰度版已成功回滚至版本 {}。'.format(model_id, version))
        elif 'FAILURE' == progress:
            output('模型 {} 灰度回滚至版本 {} 失败。任务结果：{}'.format(
                model_id, version, result))
        elif 'REVOKED' == progress:
            output('模型 {} 灰度回滚至版本 {} 任务已被撤销失败。'.format(model_id, version))

    def model_versions(self, args):
        limit = args.limit
        next_page = args.next_page
        home = get_home()
        project = get_active_project()
        config_path = self.get_model_config_path()
        config = get_config_parser(config_path)
        section_name = DEFAULT_MODEL_SECTION
        model_id = config_get(config.get, section_name, 'model_id')
        oauth2_section = DEFAULT_OAUTH2_SECTION
        token_section = DEFAULT_TOKEN_SECTION
        config = read_config(project)
        access_token = config_get(config.get, token_section, 'access_token')
        endpoint = config_get(config.get, oauth2_section, 'endpoint')
        api = API(
            access_token, endpoint=endpoint, timeout=DEFAULT_MODEL_TIMEOUT)
        try:
            result = api.model_versions(
                model_id, limit=limit, next_page=next_page)
        except APIError as e:
            output('获取模型版本列表失败：{}'.format(e))
            sys.exit(1)
        output('第 {} 页，模型 {} 已发布版本：'.format(next_page or 1, model_id))
        output('...')
        length = None
        items = result['result']
        for item in items:
            version = item['version']
            message = item['message']
            create_time = datetime.fromtimestamp(item['create_time'])
            if length is None:
                length = len(str(version))
            output('发布时间：{}，版本号：{}，发布记录：{}'.format(
                create_time, str(version).rjust(length), message))
        output('...')
        size = len(items)
        if size == 0 or size < limit or next_page == result['next_page']:
            output('下一页：无')
        else:
            output('下一页：{}'.format(result['next_page']))

    def run_model(self, args):
        params = {}
        if args.file:
            filepath = args.file
            if not exists(filepath):
                output('文件不存在：{}'.format(filepath))
                sys.exit(1)
            if isdir(filepath):
                output('存在文件夹：{}'.format(filepath))
                sys.exit(1)
            with open(filepath, 'r') as fp:
                params = json.load(fp)
        elif args.args:
            params = dict(args.args)
        draft = args.draft
        test = args.test
        home = get_home()
        project = get_active_project()
        config_path = self.get_model_config_path()
        config = get_config_parser(config_path)
        section_name = DEFAULT_MODEL_SECTION
        model_id = config_get(config.get, section_name, 'model_id')
        oauth2_section = DEFAULT_OAUTH2_SECTION
        token_section = DEFAULT_TOKEN_SECTION
        config = read_config(project)
        access_token = config_get(config.get, token_section, 'access_token')
        endpoint = config_get(config.get, oauth2_section, 'endpoint')
        api = API(
            access_token, endpoint=endpoint, timeout=DEFAULT_MODEL_TIMEOUT)
        try:
            if draft is True:
                result = api.invoke_draft_model(model_id, **params)
            elif test is True:
                api = API(
                    access_token, endpoint=TEST_SERVER_ENDPOINT,
                    timeout=DEFAULT_MODEL_TIMEOUT)
                result = api.invoke_model(model_id, **params)
            else:
                result = api.invoke_model(model_id, **params)
        except APIError as e:
            output('模型运行失败：{}'.format(e))
            sys.exit(1)
        output('模型返回值：')
        output(json.dumps(result.json(), indent=4, ensure_ascii=False))

    def get_model_config_path(self, home=None):
        if home is None:
            home = get_home()
        config_dir = join(home, '.bge')
        model_config_path = join(home, 'model.ini')
        if not exists(model_config_path) or not exists(config_dir):
            output('请确认当前目录是否为 BGE 开放平台解读模型脚手架项目的根目录。')
            sys.exit(1)
        return model_config_path

    def _read_model_project(self, args):
        config_path = self.get_model_config_path()
        output('配置文件：{}'.format(config_path))
        output('配置详情：')
        output('')
        with open(config_path, 'r') as config_file:
            output(config_file.read())

    def install_deps(self, args):
        package_name = args.package_name
        requirements = args.requirements
        upgrade = args.upgrade
        force_reinstall = args.force_reinstall
        pkgs = ' '.join(package_name)
        home = get_home()
        config_path = self.get_model_config_path()
        config = get_config_parser(config_path)
        section_name = DEFAULT_MODEL_SECTION
        model_id = config_get(config.get, section_name, 'model_id')
        runtime = config_get(config.get, section_name, 'runtime')
        image_name = RUNTIMES[runtime]
        container_name = generate_container_name(model_id)
        deps = []
        if pkgs:
            deps.append(pkgs)
        if requirements:
            deps.append('-r {}'.format(requirements))
        command = 'whereis docker' if SYS_STR == 'windows' else 'which docker'
        with os.popen(command) as f:
            docker_path = f.read()
        if not docker_path:
            output(
                '请先安装 docker，参考 https://docs.docker.com/engine/install/')
            sys.exit(1)
        try:
            client = docker.from_env()
        except docker.errors.DockerException:
            output('请确认 docker 服务是否已开启。')
            output('启动命令：/etc/init.d/docker start')
            sys.exit(1)
        command = ['pip install']
        if upgrade:
            command.append('--upgrade')
        if force_reinstall:
            command.append('--force-reinstall')
        command.append('-t /code/lib {}'.format(' '.join(deps)))
        command = ' '.join(command)
        try:
            client.images.get(image_name)
        except docker.errors.NotFound:
            output('本地 docker 镜像 {} 不存在，开始拉取...'.format(image_name))
            return_code = os.system('docker pull {}'.format(image_name))
            if return_code != 0:
                output('镜像拉取失败，请重试')
                sys.exit(1)
            output('镜像拉取成功')
        output('开始安装模型依赖包...')
        command = ('sh -c "{}"').format(command)
        user = get_sys_user()
        try:
            container = client.containers.get(container_name)
            if container.status != 'exited':
                try:
                    container.remove(force=True)
                except:
                    pass
        except docker.errors.NotFound:
            pass
        container = client.containers.run(
            image_name,
            command=command,
            name=container_name,
            volumes={home: { 'bind': WORKDIR, 'mode': 'rw' }},
            stop_signal='SIGINT',
            user=user,
            detach=True,
            auto_remove=True)
        try:
            logs = container.logs(stream=True)
            for log in logs:
                sys.stdout.write(log.decode('utf-8'))
                sys.stdout.flush()
        finally:
            if container.status != 'exited':
                try:
                    container.remove(force=True)
                except:
                    pass
        output('安装完成')

    def start_model(self, args):
        output('Model debug server is starting ...')
        home = get_home()
        config_path = self.get_model_config_path()
        config = get_config_parser(config_path)
        section_name = DEFAULT_MODEL_SECTION
        model_id = config_get(config.get, section_name, 'model_id')
        runtime = config_get(config.get, section_name, 'runtime')
        image_name = RUNTIMES[runtime]
        container_name = generate_container_name(model_id)
        command = 'whereis docker' if SYS_STR == 'windows' else 'which docker'
        with os.popen(command) as f:
            docker_path = f.read()
        if not docker_path:
            output(
                '请先安装 docker，参考 https://docs.docker.com/engine/install/')
            sys.exit(1)
        try:
            client = docker.from_env()
        except docker.errors.DockerException:
            output('请确认 docker 服务是否已开启。')
            output('启动命令：/etc/init.d/docker start')
            sys.exit(1)
        try:
            client.images.get(image_name)
        except docker.errors.NotFound:
            output('本地 docker 镜像 {} 不存在，开始拉取...'.format(image_name))
            return_code = os.system('docker pull {}'.format(image_name))
            if return_code != 0:
                output('镜像拉取失败，请重试')
                sys.exit(1)
            output('镜像拉取成功')
        command = 'python -u /server/app.py'
        user = get_sys_user()
        try:
            container = client.containers.get(container_name)
            if container.status != 'exited':
                try:
                    container.remove(force=True)
                except:
                    pass
        except docker.errors.NotFound:
            pass
        container = client.containers.run(
            image_name,
            command=command,
            name=container_name,
            volumes={home: { 'bind': WORKDIR, 'mode': 'rw' }},
            stop_signal='SIGINT',
            ports={TEST_SERVER_PORT: TEST_SERVER_PORT},
            user=user,
            detach=True,
            stream=True,
            auto_remove=True)
        output('Model {} was registered'.format(model_id))
        output('\n\tURL: {}/model/{}'.format(TEST_SERVER_ENDPOINT, model_id))
        output('\tMethod: GET\n')
        try:
            logs = container.logs(stream=True, follow=True)
            for log in logs:
                sys.stdout.write(log.decode('utf-8'))
                sys.stdout.flush()
        finally:
            if container.status != 'exited':
                try:
                    container.remove(force=True)
                except:
                    pass



import posixpath
import sys

from bgesdk.client import API
from bgesdk.error import APIError
from bgesdk.management import constants
from bgesdk.management.command import BaseCommand
from bgesdk.management.utils import (
    get_active_project, config_get, read_config, output
)
from bgesdk.version import __version__


DEFAULT_OAUTH2_SECTION = constants.DEFAULT_OAUTH2_SECTION
DEFAULT_TOKEN_SECTION = constants.DEFAULT_TOKEN_SECTION


class Command(BaseCommand):

    order = 12
    help='下载文件。'

    def add_arguments(self, parser):
        parser.add_argument(
            'object_name',
            help='OSS 对象名。'
        )
        parser.add_argument(
            '-m',
            '--mode',
            default='wb',
            choices=('w', 'wb'),
            type=str,
            help='下载文件打开模式。'
        )
        parser.add_argument(
            '-r',
            '--region',
            default='domestic',
            choices=('domestic', 'international'),
            type=str,
            help='区域。'
        )
        parser.add_argument(
            '-c',
            '--chunk_size',
            type=int,
            default=8192,
            help='下载块大小。'
        )
        parser.add_argument(
            '-e',
            '--expiration_time',
            type=int,
            default=600,
            help='下载地址过期时间。'
        )
        parser.add_argument(
            '-t',
            '--access_token',
            type=str,
            help='访问令牌，为空时默认使用 bge token 命令获取到的 access_token。'
        )

    def handler(self, args):
        access_token = args.access_token
        object_name = args.object_name
        mode = args.mode
        project = get_active_project()
        oauth2_section = DEFAULT_OAUTH2_SECTION
        token_section = DEFAULT_TOKEN_SECTION
        config = read_config(project)
        if not access_token:
            access_token = config_get(config.get, token_section, 'access_token')
        endpoint = config_get(config.get, oauth2_section, 'endpoint')
        api = API(access_token, endpoint=endpoint, timeout=18.)
        filename = posixpath.split(object_name)[1]
        if filename == '':
            output('下载失败，object_name 不是文件：{}'.format(repr(object_name)))
            sys.exit(1)
        if posixpath.exists(filename):
            basename, suffix = posixpath.splitext(filename)
            i = 1
            while True:
                new_basename = '{} ({})'.format(basename, i)
                filename = ''.join((new_basename, suffix))
                if not posixpath.exists(filename):
                    break
                i += 1
        try:
            with open(filename, mode) as fp:
                api.download(
                    object_name,
                    fp,
                    region=args.region,
                    expiration_time=args.expiration_time,
                    chunk_size=args.chunk_size
                )
        except APIError as e:
            output('\n\n请求失败：{}'.format(e))
            sys.exit(1)
        output('\n\n文件下载成功：{}'.format(filename))




import argparse
import sys

from importlib import import_module

try:
    import readline
except ImportError:
    pass

from . import version
from .management.command import BaseCommand
from .management.utils import get_commands


def init_parser():
    parser = argparse.ArgumentParser(
        description=('BGE 开放平台 SDK 命令行工具提供了初始化模型脚手架、部署模型、'
                     '初始化模型文档配置文件、上传图片、部署模型文档等命令。'),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        '-V',
        '--version',
        action='version',
        help='显示当前 BGE 开放平台 Python SDK 版本号。',
        version='version {}'.format(version.__version__)
    )
    subparsers = parser.add_subparsers(
        dest='command',
        help='SDK 命令行工具可选子命令。'
    )
    command_names = get_commands()
    sort_items = []
    for sub_command_name in command_names:
        module = import_module(
            'bgesdk.management.commands.{}'.format(sub_command_name)
        )
        command_klass = getattr(module, 'Command', None)
        if command_klass and not issubclass(command_klass, BaseCommand):
            continue
        order = int(getattr(command_klass, 'order', 1))
        sort_items.append((order, sub_command_name, command_klass))
    sort_items.sort()
    for _, sub_command_name, command_klass in sort_items:
        command_klass(subparsers, sub_command_name)
    return parser


def main():
    parser = init_parser()
    if len(sys.argv) < 2:
        parser.print_help(sys.stderr)
        sys.exit(1)
    input_args = sys.argv[1:]
    args = parser.parse_args(input_args)
    try:
        args.method(args)
    except KeyboardInterrupt:
        sys.exit(1)


if '__main__' == __name__:
    main()



import os
import posixpath
import sys

from posixpath import join, isfile

from bgesdk.client import API
from bgesdk.error import APIError
from bgesdk.management import constants
from bgesdk.management.command import BaseCommand
from bgesdk.management.utils import (
    get_active_project, config_get, read_config, output
)
from bgesdk.version import __version__


DEFAULT_OAUTH2_SECTION = constants.DEFAULT_OAUTH2_SECTION
DEFAULT_TOKEN_SECTION = constants.DEFAULT_TOKEN_SECTION


class Command(BaseCommand):

    order = 11
    help='上传目录下文件（不递归上传子目录中文件）。'

    def add_arguments(self, parser):
        parser.add_argument(
            'dirpath',
            type=str,
            help='要上传的文件夹。'
        )
        parser.add_argument(
            '-i',
            '--cmk_id',
            help='阿里云 KMS 服务密钥 ID，提供时代表使用加密方式上传文件。'
        )
        parser.add_argument(
            '-t',
            '--access_token',
            type=str,
            help='访问令牌，为空时默认使用 bge token 命令获取到的 access_token。'
        )

    def handler(self, args):
        access_token = args.access_token
        project = get_active_project()
        oauth2_section = DEFAULT_OAUTH2_SECTION
        token_section = DEFAULT_TOKEN_SECTION
        config = read_config(project)
        if not access_token:
            access_token = config_get(config.get, token_section, 'access_token')
        endpoint = config_get(config.get, oauth2_section, 'endpoint')
        api = API(access_token, endpoint=endpoint, timeout=18.)
        dirpath = args.dirpath
        cmk_id = args.cmk_id
        files = []
        for filename in os.listdir(dirpath):
            filepath = join(dirpath, filename)
            if isfile(filepath):
                files.append(filepath)
                continue
        if not files:
            output('文件夹中没有可上传的文件')
            sys.exit(1)
        try:
            object_names = api.upload_dir(dirpath, cmk_id=cmk_id)
        except APIError as e:
            output('\n\n请求失败：{}'.format(e))
            sys.exit(1)
        output('\n文件上传成功：')
        for i, object_name in enumerate(object_names, 1):
            output('{}. {}'.format(i, object_name))



#-*- coding: utf-8 -*-

import logging

from six.moves import http_client


def new_logger(name, verbose=False):
    logging.basicConfig()
    if verbose:
        http_client.HTTPConnection.debuglevel = 0
        requests_log = logging.getLogger("urllib3")
        requests_log.setLevel(logging.INFO)
        requests_log.propagate = False
        logger = logging.getLogger(name)
        logger.setLevel(logging.DEBUG)
        return logger
    else:
        logger = logging.getLogger(name)
        logger.setLevel(logging.INFO)
        return logger


def human_byte(size, dot=2):
    size = float(size)
    if size < pow(1024, 2):
        size = str(round(size / pow(1024, 1), dot)) + 'KB'
    elif pow(1024, 2) <= size < pow(1024, 3):
        size = str(round(size / pow(1024, 2), dot)) + 'MB'
    else:
        size = str(round(size / pow(1024, 3), dot)) + 'GB'
    return size



#-*- coding: utf-8 -*-

import os
import pkgutil
import platform
import six
import sys

from os.path import expanduser
from posixpath import join, exists, abspath
from six.moves import configparser, http_client

from . import __path__, constants

if six.PY2:
    ConfigParser = configparser.SafeConfigParser
else:
    ConfigParser = configparser.ConfigParser

NoSectionError = configparser.NoSectionError
NoOptionError = configparser.NoOptionError
HTTPConnection = http_client.HTTPConnection

SYS_STR = platform.system().lower()


def get_home():
    return abspath('.')


def get_user_home():
    return expanduser("~")


def get_config_dir():
    user_home = get_user_home()
    bge_home = join(user_home, '.bge')
    if not exists(bge_home):
        os.makedirs(bge_home)
    return bge_home


def get_config_path(project, check_exists=True):
    config_dir = get_config_dir()
    config_path = join(config_dir, '{}.ini'.format(project))
    if not exists(config_path) and check_exists:
        print('{} 不存在，请运行 bge config 或 bge config add 命令初始化项目配置'
              .format(project))
        sys.exit(1)
    return config_path


def confirm(prompt=None):
    if prompt is None:
        prompt = '确认'
    prompt = '%s\n[%s]/%s: ' % (prompt, 'Y', 'n')
    while True:
        ans = input(prompt)
        if ans not in ['y', 'Y', 'n', 'N']:
            print('请输入 Y 或 n')
            continue
        if ans == 'y' or ans == 'Y':
            return True
        if ans == 'n' or ans == 'N':
            return False


def find_commands(commands_dir):
    """查找子命令列表"""
    return [
        name for _, name, is_pkg in pkgutil.iter_modules(
            [commands_dir])
    ]


def get_commands():
    """获取全部子命令"""
    return find_commands(join(__path__[0], 'commands'))


def get_active_path():
    config_dir = get_config_dir()
    return join(config_dir, 'activate_project')


def get_active_project():
    active_path = get_active_path()
    project = constants.DEFAULT_PROJECT
    if exists(active_path):
        with open(active_path, 'r') as fp:
            project = fp.read()
    else:
        with open(active_path, 'w') as fp:
            fp.write(project)
    return project


def generate_container_name(model_id):
    return '{}{}'.format(constants.MODEL_CONTAINER_PREFIX, model_id)


def config_get(method, section_name, key, default=None):
    try:
        return method(section_name, key)
    except (NoOptionError, NoSectionError):
        return default


def secure_str(s):
    if len(s) > 6:
        return s.replace(s[3:-3], '*' * 6)
    if len(s) >= 1:
        return s[0] + '*' * 6
    return s


def read_config(project):
    config_path = get_config_path(project)
    config = ConfigParser()
    config.read(config_path)
    return config


def get_config_parser(path):
    config_parser = ConfigParser()
    config_parser.read(path)
    return config_parser


def output(*args, sep=' ', end='\n'):
    args = map(lambda x: str(x), args)
    content = '{}'.format(sep).join(args)
    return sys.stdout.write("{}{}".format(content, end))


def get_sys_user():
    if SYS_STR == 'windows':
        user = '1000:1000'
    else:
        import pwd
        pwd_info = pwd.getpwuid(os.getuid())
        uid = pwd_info.pw_uid
        gid = pwd_info.pw_gid
        user = '{}:{}'.format(uid, gid)
    return user



import json
import sys

from bgesdk.client import API
from bgesdk.error import APIError
from bgesdk.management import constants
from bgesdk.management.command import BaseCommand
from bgesdk.management.utils import (
    get_active_project, config_get, read_config, output
)
from bgesdk.models import ModelEncoder
from bgesdk.version import __version__


DEFAULT_OAUTH2_SECTION = constants.DEFAULT_OAUTH2_SECTION
DEFAULT_TOKEN_SECTION = constants.DEFAULT_TOKEN_SECTION


class Command(BaseCommand):

    order = 8
    help='请求数据流。'

    def add_arguments(self, parser):
        parser.add_argument(
            'data_element_id',
            type=str,
            help='数据元编号。'
        )
        parser.add_argument(
            '-b',
            '--biosample_id',
            type=str,
            help='生物样品编号。'
        )
        parser.add_argument(
            '--start_time',
            type=str,
            help='数据元编号。'
        )
        parser.add_argument(
            '--end_time',
            type=str,
            help='数据元编号。'
        )
        parser.add_argument(
            '--sort_direction',
            type=str,
            default='desc',
            choices=['desc', 'asc'],
            help='排序方向'
        )
        parser.add_argument(
            '-p',
            '--next_page',
            type=str,
            help='要获取的页码。'
        )
        parser.add_argument(
            '-n',
            '--limit',
            type=int,
            help='每页返回数量，默认值为 100。'
        )
        parser.add_argument(
            '--pretty',
            default=False,
            action='store_true',
            help='打印可读性高的 JSON 字符串。'
        )
        parser.add_argument(
            '-t',
            '--access_token',
            type=str,
            help='访问令牌，为空时默认使用 bge token 命令获取到的 access_token。'
        )

    def handler(self, args):
        pretty = args.pretty
        access_token = args.access_token
        data_element_id = args.data_element_id
        project = get_active_project()
        oauth2_section = DEFAULT_OAUTH2_SECTION
        token_section = DEFAULT_TOKEN_SECTION
        config = read_config(project)
        if not access_token:
            access_token = config_get(config.get, token_section, 'access_token')
        endpoint = config_get(config.get, oauth2_section, 'endpoint')
        api = API(access_token, endpoint=endpoint, timeout=18.)
        try:
            result = api.get_range_stream(
                data_element_id,
                biosample_id=args.biosample_id.upper(),
                start_time=args.start_time,
                end_time=args.end_time,
                sort_direction=args.sort_direction,
                next_page=args.next_page,
                limit=args.limit
            )
        except APIError as e:
            output('请求失败：{}'.format(e))
            sys.exit(1)
        if pretty:
            result = json.dumps(
                result, ensure_ascii=False, indent=4, cls=ModelEncoder
            )
        else:
            result = json.dumps(result, ensure_ascii=False, cls=ModelEncoder)
        output('请求成功，返回值：')
        output(result)



import argparse
import json
import os
import sys

from datetime import datetime
from posixpath import join, exists

from bgesdk.client import API
from bgesdk.error import APIError
from bgesdk.management.command import BaseCommand
from bgesdk.management.constants import (
    TAB_CHOICES, TITLE_NAME, API_TABLE, DEFAULT_TOKEN_SECTION,
    DEFAULT_OAUTH2_SECTION, DEFAULT_MODEL_TIMEOUT
)
from bgesdk.management.utils import (
    config_get, get_active_project, read_config, get_home,
    output, SYS_STR
)
from bgesdk.management.validate import validator_doc


class Command(BaseCommand):

    order = 6
    help = '对模型文档进行预览和发布，可访问 ' \
           'https://api.bge.genomics.cn/doc/scripts/model_doc.json 下载示例文件'

    def add_arguments(self, parser):
        home = get_home()
        doc_ps = parser.add_subparsers(
            help='对模型文档进行预览和发布',
            required=False
        )

        init_p = doc_ps.add_parser(
            'init',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            help='初始化 docsify 项目。'
        )
        init_p.add_argument(
            'name',
            type=str,
            default='docs',
            help='docsify 项目名称'
        )
        init_p.add_argument(
            '--home',
            type=str,
            default=home,
            help='docsify 项目生成的父级目录，默认为当前目录'
        )
        init_p.set_defaults(method=self.init_docsify, parser=init_p)

        pre_p = doc_ps.add_parser(
            'preview',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            help='对模型文档进行预览'
        )

        pre_p.add_argument(
            'path',
            type=str,
            help='模型文档的 json 文件路径。'
        )
        pre_p.add_argument(
            '--home',
            type=str,
            default=home,
            help='docsify 项目的根目录，默认为当前目录'
        )

        pre_p.set_defaults(method=self.preview, parser=pre_p)

        release_p = doc_ps.add_parser(
            'release',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            help='发布模型文档'
        )
        release_p.set_defaults(method=self.release_doc, parser=release_p)

    def handler(self, args):
        """打印 subparser 帮助信息"""
        parser = args.parser
        parser.print_help(sys.stderr)

    def init_docsify(self, args):
        name = args.name
        home = args.home
        if home is None:
            home = get_home()
        docs_dir = join(home, name)
        if exists(docs_dir):
            output('错误！{} 已存在 '.format(docs_dir))
            sys.exit(1)
        if not exists(home):
            output('错误！无法找到 home 目录 {}。'.format(home))
            sys.exit(1)
        with os.popen('docsify init {}'.format(docs_dir)) as f:
            content = f.read()
        if content:
            output('docsify 项目已初始化，路径为：{}'.format(docs_dir))
            output('请跳转至项目目录下。')

    def get_docs_dir(self, home=None):
        if home is None:
            home = get_home()
        doc_path = join(home, 'index.html')
        if not exists(doc_path):
            output('请确认当前目录或者所输入的项目路径是否为 docsify 项目的根目录。')
            sys.exit(1)
        return home

    def preview(self, args):
        command = 'whereis docsify' if SYS_STR == 'windows' else 'which docsify'
        with os.popen(command) as f:
            content = f.read()
        if not content:
            output('请先安装 docsify，参考 https://docsify.js.org/#/quickstart')
            sys.exit(1)
        path = args.path
        if not exists(path):
            output('文件路径：{} 有误，请检查。'.format(path))
            sys.exit(1)
        docs_dir = self.get_docs_dir(home=args.home)
        doc_data = json.load(open(path))
        result = validator_doc(doc_data)
        if result['valid'] is False:
            output('文件内容有误，错误内容：{}'.format(result['errors']))
            sys.exit(1)
        doc_tab = doc_data['doc_tab']
        model_id = doc_data['model_id']
        doc_content = doc_data['doc_content']
        sidebar_path = join(docs_dir, '_sidebar.md')
        file_dir = join(docs_dir, 'model_center')
        sidebar_lines = []
        req_path = 'model_center/{}.md'.format(model_id)
        for content in doc_content:
            language = content['language']
            doc_name = content['doc_name']
            if language == 'en':
                file_dir = join(
                    docs_dir, 'en', 'model_center')
                req_path = 'en/model_center/{}.md'.format(model_id)
            if not exists(file_dir):
                os.makedirs(file_dir)
            file_path = join(
                file_dir, '{}.md'.format(model_id))
            sidebar = '        * [{}]({})'.format(doc_name, req_path)
            sidebar_lines.append(sidebar)
            self._make_doc(content, file_path, model_id)

        self._write_to_sidebar(doc_tab, sidebar_path, sidebar_lines)
        self._write_index(docs_dir)
        os.system('docsify serve {} --port 3000'.format(docs_dir))

    def _make_doc(self, content, file_path, model_id):
        line_feed = '\n'
        lines = []
        divid_line = '--------'
        developer = 'Developer: {}'.format(content.get('developer', ''))
        content_title = '# {}'.format(content.get('content_title', ''))
        ctime = 'Ctime: {}'.format(
            datetime.strftime(datetime.now(), '%Y-%m-%d %H:%M:%S'))
        version_line = 'Version:'
        join_lines = [content_title, ctime, developer, version_line,
                      divid_line]
        for key, value in TITLE_NAME.items():
            title_line = '#### {}'.format(value)
            join_model_id = None
            science_detail_title = None
            if key == 'brief_intro':
                title_line = '### {}'.format(value)
                join_model_id = '**model_id**: `{}`'.format(model_id)
                science_detail_title = '### {}'.format('科学细节')
            handle_content = self._join_content(content.get(key, {}))
            join_lines.append(title_line)
            join_lines.extend(handle_content)
            join_lines.append(join_model_id)
            join_lines.append(science_detail_title)
        api_title = '### {}'.format('API调用')
        join_lines.append(api_title)

        for key, value in API_TABLE.items():

            params_line = '![{}](https://img.shields.io/badge/' \
                          '{}-{}-blue)'.format(key, value, key)
            params = dict()
            if key == 'QueryParams':
                pass
            elif key == 'Success':
                params = content.get('return_params', None)
            elif key == 'State':
                params = content.get('state_explain', None)
            params_table = self._join_table(params)
            join_lines.append(params_line)
            join_lines.append(params_table)

        example_result_line = '![Success](https://img.shields.io/badge/' \
                              '输出-Success-green)'
        json_str = '```json'
        example_result = content.get('example_result')
        example_result_json = ''
        if example_result:
            example_result_json = json.dumps(
                example_result, sort_keys=False, indent=4)
        last_json_str = '```'
        join_lines.append(example_result_line)
        join_lines.append(json_str)
        join_lines.append(example_result_json)
        join_lines.append(last_json_str)
        join_lines.append(divid_line)

        ref_line = '### {}'.format('参考文献')
        join_lines.append(ref_line)
        refs = content.get('ref', [])
        ref_lines = []
        if refs:
            for index, ref in enumerate(refs):
                ref_str = r'\[{}]: {}{}'.format(index + 1, ref, '<br>')
                ref_lines.append(ref_str)
        join_lines.append(ref_lines)
        for line in join_lines:
            if isinstance(line, list):
                for l in line:
                    lines.append(l + line_feed)
            elif isinstance(line, str):
                lines.append(line + line_feed)
            lines.append(line_feed)

        with open(file_path, 'w') as f:
            f.writelines(lines)

    def _join_content(self, content):
        line_feed = '\n'
        templates = content.get('templates', None)
        arguments = content.get('arguments', None)
        if templates:
            description_lines = []
            params = arguments.get('image', None)
            target_array = 0
            for tem in templates:
                tem = str(tem)
                word = '{%s}' % 'image'
                if word in tem:
                    word_count = tem.count(word)
                    for index in range(target_array,
                                       target_array + word_count):
                        try:
                            data = params[index]
                        except:
                            output(
                                'arguments中的关键词数组应大于'
                                '等于templates中的关键词数组')
                            sys.exit(1)
                        image_md = '''{}<br><center><img src="{}" 
                        width="386" height="386" /><div style="color: #999;
                        ">{}</div></center><br>{}'''.format(
                            line_feed, data['uri'], data['caption'],
                            line_feed)
                        tem = tem.replace(word, image_md, 1)
                    target_array += word_count
                description_lines.append(tem)
            return description_lines
        return ['未提及', 'Not mentioned']

    def _join_table(self, params):
        table_lines = []
        table_1 = '| {} | {} |'.format('数据名'.ljust(29), '描述'.ljust(29))
        table_2 = '| {} | {} |'.format('-' * 31, '-' * 31)
        table_lines.append(table_1)
        table_lines.append(table_2)
        for key, value in params.items():
            param_line = '| {} | {} |'.format(
                str(key).ljust(31), str(value))
            table_lines.append(param_line)
        return table_lines

    def _write_to_sidebar(self, doc_tab, sidebar_path, sidebar_lines):
        sidebar_title = '* :chart_with_upwards_trend: 模型中心'
        tab_choices = dict(TAB_CHOICES)
        tab = '    * {}'.format(tab_choices.get(doc_tab))
        lines = [sidebar_title, tab]
        lines.extend(sidebar_lines)
        with open(sidebar_path, 'w') as f:
            for line in lines:
                f.write(line)
                f.write('\n')

    def _write_index(self, docs_dir):
        index_path = join(docs_dir, 'index.html')
        if exists(index_path) is False:
            output('docs 目录下无 index.html')
            sys.exit(1)
        lines = []
        with open(index_path, 'r') as f:
            for line in f.readlines():
                lines.append(line)
                if line == '    window.$docsify = {\n':
                    lines.append('      loadSidebar: true,\n')
                elif line == '  </script>\n':
                    lines.append(
                        '  <script src="//cdn.jsdelivr.net/npm/prismjs/'
                        'components/prism-json.min.js"></script>\n')
                    lines.append(
                        '  <script src="//cdn.jsdelivr.net/npm/docsify-kate'
                        'x@latest/dist/docsify-katex.js"></script>\n')
                    lines.append(
                        '  <link rel="stylesheet" href="//cdn.jsdelivr.net/'
                        'npm/katex@latest/dist/katex.min.css"/>\n')
        with open(index_path, 'w') as f:
            f.writelines(lines)

    def release_doc(self, args):
        project = get_active_project()
        config = read_config(project)
        access_token = config_get(
            config.get, DEFAULT_TOKEN_SECTION, 'access_token')
        endpoint = config_get(config.get, DEFAULT_OAUTH2_SECTION, 'endpoint')
        api = API(
            access_token, endpoint=endpoint, timeout=DEFAULT_MODEL_TIMEOUT)
        input_value = input('？请输入需要发布的 json 文件路径：')
        if input_value:
            if not exists(input_value):
                output('文件路径：{} 有误，请检查。'.format(input_value))
                sys.exit(1)
            doc_data = json.load(open(input_value))
            doc_tab = doc_data['doc_tab']
            model_id = doc_data['model_id']
            doc_content = doc_data['doc_content']
            try:
                result = api.upload_model_doc(doc_tab, model_id, doc_content)
            except APIError as e:
                output('模型文档上传失败：{}'.format(e))
                sys.exit(1)
            output('模型文档上传结果：')
            output(json.dumps(result.json(), indent=4, ensure_ascii=False))

        else:
            output('请输入需要预览的文档路径')
            sys.exit(1)




import sys
import webbrowser

from six.moves import input
from time import sleep

from bgesdk.client import OAuth2
from bgesdk.error import APIError
from bgesdk.management import constants
from bgesdk.management.command import BaseCommand
from bgesdk.management.utils import (
    get_active_project, config_get, get_config_path, get_config_parser,
    output
)


class Command(BaseCommand):

    order = 3
    help='授权码模式登录并获取授权令牌。'

    def add_arguments(self, parser):
        parser.add_argument(
            'redirect_uri',
            help='授权完成后的回调地址。'
        )
        parser.add_argument(
            '--state',
            help='额外信息，原样作为参数返回给回调地址。'
        )
        parser.add_argument(
            '--scopes',
            help='权限标识'
        )

    def handler(self, args):
        project = get_active_project()
        config_path = get_config_path(project)
        config_parser = get_config_parser(config_path)
        oauth2_section = constants.DEFAULT_OAUTH2_SECTION
        client_id = config_get(config_parser.get, oauth2_section, 'client_id')
        client_secret = config_get(
            config_parser.get, oauth2_section, 'client_secret')
        endpoint = config_get(config_parser.get, oauth2_section, 'endpoint')
        redirect_uri = args.redirect_uri
        oauth2 = OAuth2(
            client_id, client_secret, endpoint=endpoint, timeout=60.)
        authorization_url = oauth2.get_authorization_url(
            redirect_uri, state=args.state, scopes=args.scopes
        )
        output('授权页面地址为：{}'.format(authorization_url))
        output('浏览器启动中...')
        webbrowser.open(authorization_url, new=1, autoraise=False)
        sleep(1)
        output('\n请在浏览器登录并完成授权，复制跳转后页面链接的 code 参数并关闭浏览器。\n')
        while True:
            code = input('请输入 code：').strip()
            if code:
                break
        try:
            token_result = oauth2.exchange_authorization_code(
                code, redirect_uri=redirect_uri
            )
        except APIError as e:
            output('令牌获取出错: {}'.format(e))
            sys.exit(1)
        self._write_token_config(project, token_result)
        output('令牌内容如下：')
        output('')
        for key in [
                'access_token', 'token_type', 'expires_in', 'scope',
                'refresh_token']:
            output('{} = {}'.format(key, token_result[key]))

    def _write_token_config(self, project, token_result):
        access_token = token_result['access_token']
        token_type = token_result['token_type']
        expires_in = str(token_result['expires_in'])
        scope = token_result['scope']
        refresh_token = token_result['refresh_token']
        config_path = get_config_path(project)
        config_parser = get_config_parser(config_path)
        section_name = constants.DEFAULT_TOKEN_SECTION
        if section_name not in config_parser.sections():
            config_parser.add_section(section_name)
        config_parser.set(section_name, 'access_token', access_token)
        config_parser.set(section_name, 'token_type', token_type)
        config_parser.set(section_name, 'expires_in', expires_in)
        config_parser.set(section_name, 'scope', scope)
        config_parser.set(section_name, 'refresh_token', refresh_token)
        with open(config_path, 'w') as config_file:
            config_parser.write(config_file)
        output('令牌已保存至：{}'.format(config_path))
